"""
This module contains metrics for usage with the Hugging Face Trainer.

IMPORTANT:
- This is a shared file between OrcaLib and the Orca SDK.
- Please ensure that it does not have any dependencies on the OrcaLib code.
- Make sure to edit this file in orcalib/shared and NOT in orca_sdk, since it will be overwritten there.

"""

from dataclasses import dataclass
from typing import Any, Literal, TypedDict, cast

import numpy as np
import sklearn.metrics
from numpy.typing import NDArray


# we don't want to depend on scipy or torch in orca_sdk
def softmax(logits: np.ndarray, axis: int = -1) -> np.ndarray:
    shifted = logits - np.max(logits, axis=axis, keepdims=True)
    exps = np.exp(shifted)
    return exps / np.sum(exps, axis=axis, keepdims=True)


# We don't want to depend on transformers just for the eval_pred type in orca_sdk
def transform_eval_pred(eval_pred: Any) -> tuple[NDArray, NDArray[np.float32]]:
    # convert results from Trainer compute_metrics param for use in calculate_classification_metrics
    logits, references = eval_pred  # transformers.trainer_utils.EvalPrediction
    if isinstance(logits, tuple):
        logits = logits[0]
    if not isinstance(logits, np.ndarray):
        raise ValueError("Logits must be a numpy array")
    if not isinstance(references, np.ndarray):
        raise ValueError(
            "Multiple label columns found, use the `label_names` training argument to specify which one to use"
        )

    return (references, logits)


class PRCurve(TypedDict):
    thresholds: list[float]
    precisions: list[float]
    recalls: list[float]


def calculate_pr_curve(
    references: NDArray[np.int64],
    probabilities: NDArray[np.float32],
    max_length: int = 100,
) -> PRCurve:
    if probabilities.ndim == 1:
        probabilities_slice = probabilities
    elif probabilities.ndim == 2:
        probabilities_slice = probabilities[:, 1]
    else:
        raise ValueError("Probabilities must be 1 or 2 dimensional")

    if len(probabilities_slice) != len(references):
        raise ValueError("Probabilities and references must have the same length")

    precisions, recalls, thresholds = sklearn.metrics.precision_recall_curve(references, probabilities_slice)

    # Convert all arrays to float32 immediately after getting them
    precisions = precisions.astype(np.float32)
    recalls = recalls.astype(np.float32)
    thresholds = thresholds.astype(np.float32)

    # Concatenate with 0 to include the lowest threshold
    thresholds = np.concatenate(([0], thresholds))

    # Sort by threshold
    sorted_indices = np.argsort(thresholds)
    thresholds = thresholds[sorted_indices]
    precisions = precisions[sorted_indices]
    recalls = recalls[sorted_indices]

    if len(precisions) > max_length:
        new_thresholds = np.linspace(0, 1, max_length, dtype=np.float32)
        new_precisions = np.interp(new_thresholds, thresholds, precisions)
        new_recalls = np.interp(new_thresholds, thresholds, recalls)
        thresholds = new_thresholds
        precisions = new_precisions
        recalls = new_recalls

    return PRCurve(
        thresholds=cast(list[float], thresholds.tolist()),
        precisions=cast(list[float], precisions.tolist()),
        recalls=cast(list[float], recalls.tolist()),
    )


class ROCCurve(TypedDict):
    thresholds: list[float]
    false_positive_rates: list[float]
    true_positive_rates: list[float]


def calculate_roc_curve(
    references: NDArray[np.int64],
    probabilities: NDArray[np.float32],
    max_length: int = 100,
) -> ROCCurve:
    if probabilities.ndim == 1:
        probabilities_slice = probabilities
    elif probabilities.ndim == 2:
        probabilities_slice = probabilities[:, 1]
    else:
        raise ValueError("Probabilities must be 1 or 2 dimensional")

    if len(probabilities_slice) != len(references):
        raise ValueError("Probabilities and references must have the same length")

    # Convert probabilities to float32 before calling sklearn_roc_curve
    probabilities_slice = probabilities_slice.astype(np.float32)
    fpr, tpr, thresholds = sklearn.metrics.roc_curve(references, probabilities_slice)

    # Convert all arrays to float32 immediately after getting them
    fpr = fpr.astype(np.float32)
    tpr = tpr.astype(np.float32)
    thresholds = thresholds.astype(np.float32)

    # We set the first threshold to 1.0 instead of inf for reasonable values in interpolation
    thresholds[0] = 1.0

    # Sort by threshold
    sorted_indices = np.argsort(thresholds)
    thresholds = thresholds[sorted_indices]
    fpr = fpr[sorted_indices]
    tpr = tpr[sorted_indices]

    if len(fpr) > max_length:
        new_thresholds = np.linspace(0, 1, max_length, dtype=np.float32)
        new_fpr = np.interp(new_thresholds, thresholds, fpr)
        new_tpr = np.interp(new_thresholds, thresholds, tpr)
        thresholds = new_thresholds
        fpr = new_fpr
        tpr = new_tpr

    return ROCCurve(
        false_positive_rates=cast(list[float], fpr.tolist()),
        true_positive_rates=cast(list[float], tpr.tolist()),
        thresholds=cast(list[float], thresholds.tolist()),
    )


@dataclass
class ClassificationMetrics:
    f1_score: float
    """F1 score of the predictions"""

    accuracy: float
    """Accuracy of the predictions"""

    loss: float
    """Cross-entropy loss of the logits"""

    anomaly_score_mean: float | None = None
    """Mean of anomaly scores across the dataset"""

    anomaly_score_median: float | None = None
    """Median of anomaly scores across the dataset"""

    anomaly_score_variance: float | None = None
    """Variance of anomaly scores across the dataset"""

    roc_auc: float | None = None
    """Receiver operating characteristic area under the curve"""

    pr_auc: float | None = None
    """Average precision (area under the curve of the precision-recall curve)"""

    pr_curve: PRCurve | None = None
    """Precision-recall curve"""

    roc_curve: ROCCurve | None = None
    """Receiver operating characteristic curve"""

    def __repr__(self) -> str:
        return (
            "ClassificationMetrics({\n"
            + f"    accuracy: {self.accuracy:.4f},\n"
            + f"    f1_score: {self.f1_score:.4f},\n"
            + (f"    roc_auc: {self.roc_auc:.4f},\n" if self.roc_auc else "")
            + (f"    pr_auc: {self.pr_auc:.4f},\n" if self.pr_auc else "")
            + (
                f"    anomaly_score: {self.anomaly_score_mean:.4f} ± {self.anomaly_score_variance:.4f},\n"
                if self.anomaly_score_mean
                else ""
            )
            + "})"
        )


def calculate_classification_metrics(
    expected_labels: list[int] | NDArray[np.int64],
    logits: list[list[float]] | list[NDArray[np.float32]] | NDArray[np.float32],
    anomaly_scores: list[float] | None = None,
    average: Literal["micro", "macro", "weighted", "binary"] | None = None,
    multi_class: Literal["ovr", "ovo"] = "ovr",
    include_curves: bool = False,
) -> ClassificationMetrics:
    references = np.array(expected_labels)

    logits = np.array(logits)
    if logits.ndim == 1:
        if (logits > 1).any() or (logits < 0).any():
            raise ValueError("Logits must be between 0 and 1 for binary classification")
        # convert 1D probabilities (binary) to 2D logits
        logits = np.column_stack([1 - logits, logits])
        probabilities = logits  # no need to convert to probabilities
    elif logits.ndim == 2:
        if logits.shape[1] < 2:
            raise ValueError("Use a different metric function for regression tasks")
        if not (logits > 0).all():
            # convert logits to probabilities with softmax if necessary
            probabilities = softmax(logits)
        elif not np.allclose(logits.sum(-1, keepdims=True), 1.0):
            # convert logits to probabilities through normalization if necessary
            probabilities = logits / logits.sum(-1, keepdims=True)
        else:
            probabilities = logits
    else:
        raise ValueError("Logits must be 1 or 2 dimensional")

    predictions = np.argmax(probabilities, axis=-1)

    num_classes_references = len(set(references))
    num_classes_predictions = len(set(predictions))

    if average is None:
        average = "binary" if num_classes_references == 2 else "weighted"

    anomaly_score_mean = float(np.mean(anomaly_scores)) if anomaly_scores else None
    anomaly_score_median = float(np.median(anomaly_scores)) if anomaly_scores else None
    anomaly_score_variance = float(np.var(anomaly_scores)) if anomaly_scores else None

    accuracy = sklearn.metrics.accuracy_score(references, predictions)
    f1 = sklearn.metrics.f1_score(references, predictions, average=average)
    loss = sklearn.metrics.log_loss(references, probabilities)

    if num_classes_references == num_classes_predictions:
        # special case for binary classification: https://github.com/scikit-learn/scikit-learn/issues/20186
        if num_classes_references == 2:
            roc_auc = sklearn.metrics.roc_auc_score(references, logits[:, 1])
            roc_curve = calculate_roc_curve(references, logits[:, 1]) if include_curves else None
            pr_auc = sklearn.metrics.average_precision_score(references, logits[:, 1])
            pr_curve = calculate_pr_curve(references, logits[:, 1]) if include_curves else None
        else:
            roc_auc = sklearn.metrics.roc_auc_score(references, probabilities, multi_class=multi_class)
            roc_curve = None
            pr_auc = None
            pr_curve = None
    else:
        roc_auc = None
        pr_auc = None
        pr_curve = None
        roc_curve = None

    return ClassificationMetrics(
        accuracy=float(accuracy),
        f1_score=float(f1),
        loss=float(loss),
        anomaly_score_mean=anomaly_score_mean,
        anomaly_score_median=anomaly_score_median,
        anomaly_score_variance=anomaly_score_variance,
        roc_auc=float(roc_auc) if roc_auc is not None else None,
        pr_auc=float(pr_auc) if pr_auc is not None else None,
        pr_curve=pr_curve,
        roc_curve=roc_curve,
    )


@dataclass
class RegressionMetrics:
    mse: float
    """Mean squared error of the predictions"""

    rmse: float
    """Root mean squared error of the predictions"""

    mae: float
    """Mean absolute error of the predictions"""

    r2: float
    """R-squared score (coefficient of determination) of the predictions"""

    explained_variance: float
    """Explained variance score of the predictions"""

    loss: float
    """Mean squared error loss of the predictions"""

    anomaly_score_mean: float | None = None
    """Mean of anomaly scores across the dataset"""

    anomaly_score_median: float | None = None
    """Median of anomaly scores across the dataset"""

    anomaly_score_variance: float | None = None
    """Variance of anomaly scores across the dataset"""

    def __repr__(self) -> str:
        return (
            "RegressionMetrics({\n"
            + f"    mae: {self.mae:.4f},\n"
            + f"    rmse: {self.rmse:.4f},\n"
            + f"    r2: {self.r2:.4f},\n"
            + (
                f"    anomaly_score: {self.anomaly_score_mean:.4f} ± {self.anomaly_score_variance:.4f},\n"
                if self.anomaly_score_mean
                else ""
            )
            + "})"
        )


def calculate_regression_metrics(
    expected_scores: NDArray[np.float32] | list[float],
    predicted_scores: NDArray[np.float32] | list[float],
    anomaly_scores: list[float] | None = None,
) -> RegressionMetrics:
    """
    Calculate regression metrics for model evaluation.

    Params:
        references: True target values
        predictions: Predicted values from the model
        anomaly_scores: Optional anomaly scores for each prediction

    Returns:
        Comprehensive regression metrics including MSE, RMSE, MAE, R², and explained variance

    Raises:
        ValueError: If predictions and references have different lengths
    """
    references = np.array(expected_scores)
    predictions = np.array(predicted_scores)

    if len(predictions) != len(references):
        raise ValueError("Predictions and references must have the same length")

    anomaly_score_mean = float(np.mean(anomaly_scores)) if anomaly_scores else None
    anomaly_score_median = float(np.median(anomaly_scores)) if anomaly_scores else None
    anomaly_score_variance = float(np.var(anomaly_scores)) if anomaly_scores else None

    # Calculate core regression metrics
    mse = float(sklearn.metrics.mean_squared_error(references, predictions))
    rmse = float(np.sqrt(mse))
    mae = float(sklearn.metrics.mean_absolute_error(references, predictions))
    r2 = float(sklearn.metrics.r2_score(references, predictions))
    explained_var = float(sklearn.metrics.explained_variance_score(references, predictions))

    return RegressionMetrics(
        mse=mse,
        rmse=rmse,
        mae=mae,
        r2=r2,
        explained_variance=explained_var,
        loss=mse,  # For regression, loss is typically MSE
        anomaly_score_mean=anomaly_score_mean,
        anomaly_score_median=anomaly_score_median,
        anomaly_score_variance=anomaly_score_variance,
    )



import pytest

from .pagination import Page, PagedIterable


class MockEndpoint:
    """Mock paginated endpoint for testing"""

    def __init__(self, total_items: int):
        self.items = list(range(total_items))
        self.fetch_count = 0

    def fetch(self, limit: int, offset: int) -> Page[int]:
        self.fetch_count += 1
        end_index = min(offset + limit, len(self.items))
        items = self.items[offset:end_index]
        return {"items": items, "count": len(self.items)}


def test_basic_pagination():
    # Given a mock endpoint with 5 items
    endpoint = MockEndpoint(5)
    # When doing a paginated iteration
    paginated = PagedIterable(endpoint.fetch, page_size=2)
    # Then we should be able to iterate through all items
    assert list(paginated) == [0, 1, 2, 3, 4]
    # And the length should be correct
    assert len(paginated) == 5
    # And 3 requests: [0,1], [2,3], [4] should have been made, one for each page
    assert endpoint.fetch_count == 3


def test_empty_results():
    # Given an empty mock endpoint
    endpoint = MockEndpoint(0)
    # When doing a paginated iteration
    paginated = PagedIterable(endpoint.fetch, page_size=5)
    # Then we should get an empty list
    assert list(paginated) == []
    # And the length should be 0
    assert len(paginated) == 0
    # And only one request should have been made, for the first page
    assert endpoint.fetch_count == 1


def test_transform_function():
    # Given a mock endpoint with 4 items
    endpoint = MockEndpoint(4)
    # And a transform function that doubles the items
    transform = lambda x: f"2x={2*x}"
    # When doing a paginated iteration with a transform function
    paginated = PagedIterable(endpoint.fetch, transform=transform, page_size=2)
    # Then we should get the transformed items
    assert list(paginated) == ["2x=0", "2x=2", "2x=4", "2x=6"]


def test_multiple_iterations():
    # Given a mock endpoint with 5 items
    endpoint = MockEndpoint(5)
    # When we do 2 paginated iterations
    paginated = PagedIterable(endpoint.fetch, page_size=2)
    result1 = list(paginated)
    result2 = list(paginated)
    # Then we should get the same items twice
    assert result1 == result2 == [0, 1, 2, 3, 4]
    # And 6 requests should have been made, 3 for each iteration
    assert endpoint.fetch_count == 6


def test_single_page_optimization():
    # Given a mock endpoint with 5 items
    endpoint = MockEndpoint(5)
    # When doing a paginated iteration with a limit that is greater than the number of items
    paginated = PagedIterable(endpoint.fetch, page_size=10)
    # Then we should get all items
    assert list(paginated) == [0, 1, 2, 3, 4]
    # And the length should be 5
    assert len(paginated) == 5
    # And only one request should have been made
    assert endpoint.fetch_count == 1
    # And a second iteration should not make any additional requests
    assert list(paginated) == [0, 1, 2, 3, 4]
    assert endpoint.fetch_count == 1


def test_indexing():
    # Given a mock endpoint with 7 items
    endpoint = MockEndpoint(7)
    # When creating a paginated iterable with page size 3
    paginated = PagedIterable(endpoint.fetch, page_size=3)
    # Then we should be able to access items by index
    assert paginated[0] == 0
    assert paginated[2] == 2
    assert paginated[6] == 6
    # And negative indices should work
    assert paginated[-1] == 6
    # And accessing out of bounds should raise IndexError
    with pytest.raises(IndexError):
        paginated[7]
    with pytest.raises(IndexError):
        paginated[-8]
    # And transforms are applied
    assert PagedIterable(endpoint.fetch, transform=lambda x: x * 10, page_size=3)[1] == 10


def test_slicing():
    # Given a mock endpoint with 10 items
    endpoint = MockEndpoint(10)
    # When creating a paginated iterable
    paginated = PagedIterable(endpoint.fetch, page_size=3)
    # Then we should be able to slice it
    assert list(paginated[2:5]) == [2, 3, 4]
    assert list(paginated[:3]) == [0, 1, 2]
    assert list(paginated[7:]) == [7, 8, 9]
    # And negative indices should work
    assert list(paginated[:-5]) == [0, 1, 2, 3, 4, 5]
    assert list(paginated[-3:]) == [7, 8, 9]
    assert list(paginated[-5:-2]) == [5, 6, 7]
    # And empty slices should work
    assert list(paginated[5:5]) == []
    # And slicing with a start and stop that are out of bounds should raise IndexError
    with pytest.raises(IndexError):
        list(paginated[20:25])
    # And slicing with a step other than 1 should raise ValueError
    with pytest.raises(ValueError):
        list(paginated[::2])
    with pytest.raises(ValueError):
        list(paginated[1:8:3])
    with pytest.raises(ValueError):
        list(paginated[::-1])
    # And transforms are applied
    assert list(PagedIterable(endpoint.fetch, transform=lambda x: x * 10, page_size=3)[1:3]) == [10, 20]



import pickle
from dataclasses import asdict, is_dataclass
from os import PathLike
from tempfile import TemporaryDirectory
from typing import Any, cast

from datasets import Dataset
from torch.utils.data import DataLoader as TorchDataLoader
from torch.utils.data import Dataset as TorchDataset


def parse_dict_like(item: Any, column_names: list[str] | None = None) -> dict:
    if isinstance(item, dict):
        return item

    if isinstance(item, tuple):
        if column_names is not None:
            assert len(item) == len(column_names)
            return {column_names[i]: item[i] for i in range(len(item))}
        elif hasattr(item, "_fields") and all(isinstance(field, str) for field in item._fields):  # type: ignore
            return {field: getattr(item, field) for field in item._fields}  # type: ignore
        else:
            raise ValueError("For datasets that return unnamed tuples, please provide column_names argument")

    if is_dataclass(item) and not isinstance(item, type):
        return asdict(item)

    raise ValueError(f"Cannot parse {type(item)}")


def parse_batch(batch: Any, column_names: list[str] | None = None) -> list[dict]:
    if isinstance(batch, list):
        return [parse_dict_like(item, column_names) for item in batch]

    batch = parse_dict_like(batch, column_names)
    keys = list(batch.keys())
    batch_size = len(batch[keys[0]])
    for key in keys:
        if not len(batch[key]) == batch_size:
            raise ValueError(f"Batch must consist of values of the same length, but {key} has length {len(batch[key])}")
    return [{key: batch[key][idx] for key in keys} for idx in range(batch_size)]


def hf_dataset_from_torch(
    torch_data: TorchDataLoader | TorchDataset, column_names: list[str] | None = None, ignore_cache=False
) -> Dataset:
    """
    Create a HuggingFace Dataset from a PyTorch DataLoader or Dataset.

    NOTE:  It's important to ignore the cached files when testing (i.e., ignore_cache=Ture), because
    cached results can ignore changes you've made to tests. This can make a test appear to succeed
    when it's actually broken or vice versa.

    Params:
        torch_data: A PyTorch DataLoader or Dataset object to create the HuggingFace Dataset from.
        column_names: Optional list of column names to use for the dataset. If not provided,
            the column names will be inferred from the data.
        ignore_cache: If True, the dataset will not be cached on disk.
    Returns:
        A HuggingFace Dataset object containing the data from the PyTorch DataLoader or Dataset.
    """
    if isinstance(torch_data, TorchDataLoader):
        dataloader = torch_data
    else:
        dataloader = TorchDataLoader(torch_data, batch_size=1, collate_fn=lambda x: x)

    def generator():
        for batch in dataloader:
            yield from parse_batch(batch, column_names=column_names)

    if ignore_cache:
        with TemporaryDirectory() as temp_dir:
            ds = Dataset.from_generator(generator, cache_dir=temp_dir)
    else:
        ds = Dataset.from_generator(generator)

    if not isinstance(ds, Dataset):
        raise ValueError(f"Failed to create dataset from generator: {type(ds)}")
    return ds


def hf_dataset_from_disk(file_path: str | PathLike) -> Dataset:
    """
    Load a dataset from disk into a HuggingFace Dataset object.

    Params:
        file_path: Path to the file on disk to create the memoryset from. The file type will
                be inferred from the file extension. The following file types are supported:

                - .pkl: [`Pickle`][pickle] files containing lists of dictionaries or dictionaries of columns
                - .json/.jsonl: [`JSON`][json] and [`JSON`] Lines files
                - .csv: [`CSV`][csv] files
                - .parquet: [`Parquet`](https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetFile.html#pyarrow.parquet.ParquetFile) files
                - dataset directory: Directory containing a saved HuggingFace [`Dataset`][datasets.Dataset]

    Returns:
        A HuggingFace Dataset object containing the loaded data.

    Raises:
        [`ValueError`][ValueError]: If the pickle file contains unsupported data types or if
            loading the dataset fails for any reason.
    """
    if str(file_path).endswith(".pkl"):
        data = pickle.load(open(file_path, "rb"))
        if isinstance(data, list):
            return Dataset.from_list(data)
        elif isinstance(data, dict):
            return Dataset.from_dict(data)
        else:
            raise ValueError(f"Unsupported pickle file: {file_path}")
    elif str(file_path).endswith(".json"):
        hf_dataset = Dataset.from_json(file_path)
    elif str(file_path).endswith(".jsonl"):
        hf_dataset = Dataset.from_json(file_path)
    elif str(file_path).endswith(".csv"):
        hf_dataset = Dataset.from_csv(file_path)
    elif str(file_path).endswith(".parquet"):
        hf_dataset = Dataset.from_parquet(file_path)
    else:
        try:
            hf_dataset = Dataset.load_from_disk(file_path)
        except Exception as e:
            raise ValueError(f"Failed to load dataset from disk: {e}")

    return cast(Dataset, hf_dataset)



from __future__ import annotations

import logging
import os
import tempfile
import zipfile
from datetime import datetime
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import Literal, Union, cast

import pandas as pd
import pyarrow as pa
from datasets import Dataset, DatasetDict
from httpx._types import FileTypes  # type: ignore
from pyarrow import parquet
from torch.utils.data import DataLoader as TorchDataLoader
from torch.utils.data import Dataset as TorchDataset
from tqdm.auto import tqdm

from ._utils.common import CreateMode, DropMode
from ._utils.data_parsing import hf_dataset_from_torch
from ._utils.tqdm_file_reader import TqdmFileReader
from .client import DatasourceMetadata, orca_api


def _upload_files_to_datasource(
    name: str,
    file_paths: list[Path],
    description: str | None = None,
) -> DatasourceMetadata:
    """
    Helper function to upload files to create a datasource using manual HTTP requests.

    This bypasses the generated client because it doesn't handle file uploads properly.

    Params:
        name: Name for the datasource
        file_paths: List of file paths to upload
        description: Optional description for the datasource

    Returns:
        Metadata for the created datasource
    """
    files: list[tuple[Literal["files"], FileTypes]] = []

    # Calculate total size for all files
    total_size = sum(file_path.stat().st_size for file_path in file_paths)

    with tqdm(total=total_size, unit="B", unit_scale=True, desc="Uploading") as pbar:
        for file_path in file_paths:
            buffered_reader = open(file_path, "rb")
            tqdm_reader = TqdmFileReader(buffered_reader, pbar)
            files.append(("files", (file_path.name, cast(bytes, tqdm_reader))))

        # Use manual HTTP request for file uploads
        metadata = orca_api.POST(
            "/datasource/upload",
            files=files,
            data={"name": name, "description": description},
        )

    return metadata


def _handle_existing_datasource(name: str, if_exists: CreateMode) -> Union["Datasource", None]:
    """
    Helper function to handle the common pattern of checking if a datasource exists
    and taking action based on the if_exists parameter.

    Params:
        name: Name of the datasource to check
        if_exists: What to do if a datasource with the same name already exists

    Returns:
        Datasource instance if opening existing, None if should proceed with creation

    Raises:
        ValueError: If the datasource already exists and if_exists is "error"
    """
    if Datasource.exists(name):
        if if_exists == "error":
            raise ValueError(f"Dataset with name {name} already exists")
        elif if_exists == "open":
            return Datasource.open(name)
    return None


class Datasource:
    """
    A Handle to a datasource in the OrcaCloud

    A Datasource is a collection of data saved to the OrcaCloud that can be used to create a [`Memoryset`][orca_sdk.LabeledMemoryset].
    It can be created from a Hugging Face Dataset, a PyTorch DataLoader or Dataset, a list of dictionaries, a dictionary of columns, a pandas DataFrame, a pyarrow Table, or a local file.

    Attributes:
        id: Unique identifier for the datasource
        name: Unique name of the datasource
        description: Optional description of the datasource
        length: Number of rows in the datasource
        created_at: When the datasource was created
        columns: Dictionary of column names and types
    """

    id: str
    name: str
    description: str | None
    length: int
    created_at: datetime
    updated_at: datetime
    columns: dict[str, str]

    def __init__(self, metadata: DatasourceMetadata):
        # for internal use only, do not document
        self.id = metadata["id"]
        self.name = metadata["name"]
        self.length = metadata["length"]
        self.created_at = datetime.fromisoformat(metadata["created_at"])
        self.updated_at = datetime.fromisoformat(metadata["updated_at"])
        self.description = metadata["description"]
        self.columns = {
            column["name"]: (
                f"enum({', '.join(f'{option!r}' for option in column['enum_options'] or []) if 'enum_options' in column else ''})"
                if column["type"] == "ENUM"
                else "str" if column["type"] == "STRING" else column["type"].lower()
            )
            for column in metadata["columns"]
        }

    def __eq__(self, other) -> bool:
        return isinstance(other, Datasource) and self.id == other.id

    def __repr__(self) -> str:
        return (
            "Datasource({\n"
            + f"    name: '{self.name}',\n"
            + f"    length: {self.length},\n"
            + "    columns: {{\n        "
            + "\n        ".join([f"{k}: {v}" for k, v in self.columns.items()])
            + "\n    }}\n"
            + "})"
        )

    @classmethod
    def from_hf_dataset(
        cls, name: str, dataset: Dataset, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a Hugging Face Dataset

        Params:
            name: Required name for the new datasource (must be unique)
            dataset: The Hugging Face Dataset to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        with tempfile.TemporaryDirectory() as tmp_dir:
            dataset.save_to_disk(tmp_dir)

            # Get all file paths in the directory
            file_paths = list(Path(tmp_dir).iterdir())

            # Use the helper function to upload files
            metadata = _upload_files_to_datasource(name, file_paths, description)
            return cls(metadata=metadata)

    @classmethod
    def from_hf_dataset_dict(
        cls,
        name: str,
        dataset_dict: DatasetDict,
        if_exists: CreateMode = "error",
        description: dict[str, str | None] | str | None = None,
    ) -> dict[str, Datasource]:
        """
        Create datasources from a Hugging Face DatasetDict

        Params:
            name: Name prefix for the new datasources, will be suffixed with the dataset name
            dataset_dict: The Hugging Face DatasetDict to create the datasources from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasources, can be a string or a dictionary of dataset names to descriptions

        Returns:
            A dictionary of datasource handles, keyed by the dataset name

        Raises:
            ValueError: If a datasource already exists and if_exists is `"error"`
        """
        if description is None or isinstance(description, str):
            description = {dataset_name: description for dataset_name in dataset_dict.keys()}
        return {
            dataset_name: cls.from_hf_dataset(
                f"{name}_{dataset_name}", dataset, if_exists=if_exists, description=description[dataset_name]
            )
            for dataset_name, dataset in dataset_dict.items()
        }

    @classmethod
    def from_pytorch(
        cls,
        name: str,
        torch_data: TorchDataLoader | TorchDataset,
        column_names: list[str] | None = None,
        if_exists: CreateMode = "error",
        description: str | None = None,
    ) -> Datasource:
        """
        Create a new datasource from a PyTorch DataLoader or Dataset

        Params:
            name: Required name for the new datasource (must be unique)
            torch_data: The PyTorch DataLoader or Dataset to create the datasource from
            column_names: If the provided dataset or data loader returns unnamed tuples, this
                argument must be provided to specify the names of the columns.
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        hf_dataset = hf_dataset_from_torch(torch_data, column_names=column_names)
        return cls.from_hf_dataset(name, hf_dataset, if_exists=if_exists, description=description)

    @classmethod
    def from_list(
        cls, name: str, data: list[dict], if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a list of dictionaries

        Params:
            name: Required name for the new datasource (must be unique)
            data: The list of dictionaries to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`

        Examples:
            >>> Datasource.from_list("my_datasource", [{"text": "Hello, world!", "label": 1}, {"text": "Goodbye", "label": 0}])
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        metadata = orca_api.POST(
            "/datasource",
            json={"name": name, "description": description, "content": data},
        )
        return cls(metadata=metadata)

    @classmethod
    def from_dict(
        cls, name: str, data: dict, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a dictionary of columns

        Params:
            name: Required name for the new datasource (must be unique)
            data: The dictionary of columns to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`

        Examples:
            >>> Datasource.from_dict("my_datasource", {"text": ["Hello, world!", "Goodbye"], "label": [1, 0]})
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        metadata = orca_api.POST(
            "/datasource",
            json={"name": name, "description": description, "content": data},
        )
        return cls(metadata=metadata)

    @classmethod
    def from_pandas(
        cls, name: str, dataframe: pd.DataFrame, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a pandas DataFrame

        Params:
            name: Required name for the new datasource (must be unique)
            dataframe: The pandas DataFrame to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        dataset = Dataset.from_pandas(dataframe)
        return cls.from_hf_dataset(name, dataset, if_exists=if_exists, description=description)

    @classmethod
    def from_arrow(
        cls, name: str, pyarrow_table: pa.Table, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a pyarrow Table

        Params:
            name: Required name for the new datasource (must be unique)
            pyarrow_table: The pyarrow Table to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        # Write to bytes buffer
        buffer = BytesIO()
        parquet.write_table(pyarrow_table, buffer)
        parquet_bytes = buffer.getvalue()

        metadata = orca_api.POST(
            "/datasource/upload",
            files=[("files", ("data.parquet", parquet_bytes))],
            data={"name": name, "description": description},
        )

        return cls(metadata=metadata)

    @classmethod
    def from_disk(
        cls, name: str, file_path: str | PathLike, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a local file

        Params:
            name: Required name for the new datasource (must be unique)
            file_path: Path to the file on disk to create the datasource from. The file type will
                be inferred from the file extension. The following file types are supported:

                - .pkl: [`Pickle`][pickle] files containing lists of dictionaries or dictionaries of columns
                - .json/.jsonl: [`JSON`][json] and [`JSON`] Lines files
                - .csv: [`CSV`][csv] files
                - .parquet: [`Parquet`][pyarrow.parquet.ParquetFile] files
                - dataset directory: Directory containing a saved HuggingFace [`Dataset`][datasets.Dataset]

            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        file_path = Path(file_path)

        # For dataset directories, use the upload endpoint with multiple files
        if file_path.is_dir():
            return cls.from_hf_dataset(
                name, Dataset.load_from_disk(file_path), if_exists=if_exists, description=description
            )

        # For single files, use the helper function to upload files
        metadata = _upload_files_to_datasource(name, [file_path], description)

        return cls(metadata=metadata)

    @classmethod
    def open(cls, name_or_id: str) -> Datasource:
        """
        Get a handle to a datasource by name or id in the OrcaCloud

        Params:
            name_or_id: The name or unique identifier of the datasource to get

        Returns:
            A handle to the existing datasource in the OrcaCloud

        Raises:
            LookupError: If the datasource does not exist
        """
        return cls(orca_api.GET("/datasource/{name_or_id}", params={"name_or_id": name_or_id}))

    @classmethod
    def exists(cls, name_or_id: str) -> bool:
        """
        Check if a datasource exists in the OrcaCloud

        Params:
            name_or_id: The name or id of the datasource to check

        Returns:
            `True` if the datasource exists, `False` otherwise
        """
        try:
            cls.open(name_or_id)
            return True
        except LookupError:
            return False

    @classmethod
    def all(cls) -> list[Datasource]:
        """
        List all datasource handles in the OrcaCloud

        Returns:
            A list of all datasource handles in the OrcaCloud
        """
        return [cls(metadata) for metadata in orca_api.GET("/datasource")]

    @classmethod
    def drop(cls, name_or_id: str, if_not_exists: DropMode = "error") -> None:
        """
        Delete a datasource from the OrcaCloud

        Params:
            name_or_id: The name or id of the datasource to delete
            if_not_exists: What to do if the datasource does not exist, defaults to
                `"error"`. Other options are `"ignore"` to do nothing.

        Raises:
            LookupError: If the datasource does not exist and if_not_exists is `"error"`
        """
        try:
            orca_api.DELETE("/datasource/{name_or_id}", params={"name_or_id": name_or_id})
            logging.info(f"Deleted datasource {name_or_id}")
        except LookupError:
            if if_not_exists == "error":
                raise

    def __len__(self) -> int:
        return self.length

    def download(
        self, output_dir: str | PathLike, file_type: Literal["hf_dataset", "json", "csv"] = "hf_dataset"
    ) -> None:
        """
        Download the datasource to a specified path in the specified format type

        Params:
            output_dir: The local directory where the downloaded file will be saved.
            file_type: The type of file to download.

        Returns:
            None
        """
        extension = "zip" if file_type == "hf_dataset" else file_type
        output_path = Path(output_dir) / f"{self.name}.{extension}"
        with open(output_path, "wb") as download_file:
            with orca_api.stream("GET", f"/datasource/{self.id}/download", params={"file_type": file_type}) as response:
                total_chunks = int(response.headers["X-Total-Chunks"]) if "X-Total-Chunks" in response.headers else None
                with tqdm(desc=f"Downloading", total=total_chunks, disable=total_chunks is None) as progress:
                    for chunk in response.iter_bytes():
                        download_file.write(chunk)
                        progress.update(1)

        # extract the zip file
        if extension == "zip":
            extract_dir = Path(output_dir) / self.name
            with zipfile.ZipFile(output_path, "r") as zip_ref:
                zip_ref.extractall(extract_dir)
            output_path.unlink()  # Remove the zip file after extraction
            logging.info(f"Downloaded {extract_dir}")
        else:
            logging.info(f"Downloaded {output_path}")

    def to_list(self) -> list[dict]:
        """
        Convert the datasource to a list of dictionaries.

        Returns:
            A list of dictionaries representation of the datasource.
        """
        return orca_api.GET("/datasource/{name_or_id}/download", params={"name_or_id": self.id, "file_type": "json"})



from uuid import uuid4

import numpy as np
import pytest
from datasets.arrow_dataset import Dataset

from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import ScoredMemoryset
from .regression_model import RegressionMetrics, RegressionModel
from .telemetry import RegressionPrediction


def test_create_model(regression_model: RegressionModel, scored_memoryset: ScoredMemoryset):
    assert regression_model is not None
    assert regression_model.name == "test_regression_model"
    assert regression_model.memoryset == scored_memoryset
    assert regression_model.memory_lookup_count == 3


def test_create_model_already_exists_error(scored_memoryset, regression_model: RegressionModel):
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset)
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset, if_exists="error")


def test_create_model_already_exists_return(scored_memoryset, regression_model: RegressionModel):
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset, if_exists="open", memory_lookup_count=37)

    new_model = RegressionModel.create("test_regression_model", scored_memoryset, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_regression_model"
    assert new_model.memoryset == scored_memoryset
    assert new_model.memory_lookup_count == 3


def test_create_model_unauthenticated(unauthenticated, scored_memoryset: ScoredMemoryset):
    with pytest.raises(ValueError, match="Invalid API key"):
        RegressionModel.create("test_regression_model", scored_memoryset)


def test_get_model(regression_model: RegressionModel):
    fetched_model = RegressionModel.open(regression_model.name)
    assert fetched_model is not None
    assert fetched_model.id == regression_model.id
    assert fetched_model.name == regression_model.name
    assert fetched_model.memory_lookup_count == 3
    assert fetched_model == regression_model


def test_get_model_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        RegressionModel.open("test_regression_model")


def test_get_model_invalid_input():
    with pytest.raises(ValueError, match="Invalid input"):
        RegressionModel.open("not valid id")


def test_get_model_not_found():
    with pytest.raises(LookupError):
        RegressionModel.open(str(uuid4()))


def test_get_model_unauthorized(unauthorized, regression_model: RegressionModel):
    with pytest.raises(LookupError):
        RegressionModel.open(regression_model.name)


def test_list_models(regression_model: RegressionModel):
    models = RegressionModel.all()
    assert len(models) > 0
    assert any(model.name == regression_model.name for model in models)


def test_list_models_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        RegressionModel.all()


def test_list_models_unauthorized(unauthorized, regression_model: RegressionModel):
    assert RegressionModel.all() == []


def test_update_model_attributes(regression_model: RegressionModel):
    regression_model.description = "New description"
    assert regression_model.description == "New description"

    regression_model.set(description=None)
    assert regression_model.description is None

    regression_model.set(locked=True)
    assert regression_model.locked is True

    regression_model.set(locked=False)
    assert regression_model.locked is False

    regression_model.lock()
    assert regression_model.locked is True

    regression_model.unlock()
    assert regression_model.locked is False


def test_delete_model(scored_memoryset: ScoredMemoryset):
    RegressionModel.create("regression_model_to_delete", ScoredMemoryset.open(scored_memoryset.name))
    assert RegressionModel.open("regression_model_to_delete")
    RegressionModel.drop("regression_model_to_delete")
    with pytest.raises(LookupError):
        RegressionModel.open("regression_model_to_delete")


def test_delete_model_unauthenticated(unauthenticated, regression_model: RegressionModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        RegressionModel.drop(regression_model.name)


def test_delete_model_not_found():
    with pytest.raises(LookupError):
        RegressionModel.drop(str(uuid4()))
    # ignores error if specified
    RegressionModel.drop(str(uuid4()), if_not_exists="ignore")


def test_delete_model_unauthorized(unauthorized, regression_model: RegressionModel):
    with pytest.raises(LookupError):
        RegressionModel.drop(regression_model.name)


def test_delete_memoryset_before_model_constraint_violation(hf_dataset):
    memoryset = ScoredMemoryset.from_hf_dataset("test_memoryset_delete_before_regression_model", hf_dataset)
    RegressionModel.create("test_regression_model_delete_before_memoryset", memoryset)
    with pytest.raises(RuntimeError):
        ScoredMemoryset.drop(memoryset.id)


@pytest.mark.parametrize("data_type", ["dataset", "datasource"])
def test_evaluate(
    regression_model: RegressionModel,
    eval_datasource: Datasource,
    eval_dataset: Dataset,
    data_type,
):
    """Test that model evaluation with a dataset works."""
    result = (
        regression_model.evaluate(eval_dataset)
        if data_type == "dataset"
        else regression_model.evaluate(eval_datasource)
    )

    assert isinstance(result, RegressionMetrics)
    assert np.allclose(result.mae, 0.4)
    assert 0.0 <= result.mse <= 1.0
    assert 0.0 <= result.rmse <= 1.0
    assert result.r2 is not None

    assert isinstance(result.anomaly_score_mean, float)
    assert isinstance(result.anomaly_score_median, float)
    assert isinstance(result.anomaly_score_variance, float)
    assert -1.0 <= result.anomaly_score_mean <= 1.0
    assert -1.0 <= result.anomaly_score_median <= 1.0
    assert -1.0 <= result.anomaly_score_variance <= 1.0


def test_evaluate_with_telemetry(regression_model, eval_dataset: Dataset):
    result = regression_model.evaluate(eval_dataset, record_predictions=True, tags={"test"})
    assert result is not None
    assert isinstance(result, RegressionMetrics)
    predictions = regression_model.predictions(tag="test")
    assert len(predictions) == 4
    assert all(p.tags == {"test"} for p in predictions)
    assert all(p.expected_score is not None for p in predictions)
    assert all(np.allclose(p.expected_score, s) for p, s in zip(predictions, eval_dataset["score"]))


def test_predict(regression_model: RegressionModel):
    predictions = regression_model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    assert np.allclose(predictions[0].score, 0.1)
    assert np.allclose(predictions[1].score, 0.9)
    assert 0 <= predictions[0].confidence <= 1
    assert 0 <= predictions[1].confidence <= 1


def test_predict_unauthenticated(unauthenticated, regression_model: RegressionModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        regression_model.predict(["This is excellent!", "This is terrible!"])


def test_predict_unauthorized(unauthorized, regression_model: RegressionModel):
    with pytest.raises(LookupError):
        regression_model.predict(["This is excellent!", "This is terrible!"])


def test_predict_constraint_violation(scored_memoryset: ScoredMemoryset):
    model = RegressionModel.create(
        "test_regression_model_lookup_count_too_high",
        scored_memoryset,
        memory_lookup_count=scored_memoryset.length + 2,
    )
    with pytest.raises(RuntimeError):
        model.predict("test")


def test_predict_with_prompt(regression_model: RegressionModel):
    """Test that prompt parameter is properly passed through to predictions"""
    # Test with an instruction-supporting embedding model if available
    prediction_with_prompt = regression_model.predict(
        "This product is amazing!", prompt="Represent this text for rating prediction:"
    )
    prediction_without_prompt = regression_model.predict("This product is amazing!")

    # Both should work and return valid predictions
    assert prediction_with_prompt.score is not None
    assert prediction_without_prompt.score is not None
    assert 0 <= prediction_with_prompt.confidence <= 1
    assert 0 <= prediction_without_prompt.confidence <= 1


def test_record_prediction_feedback(regression_model: RegressionModel):
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    expected_scores = [0.9, 0.1]
    regression_model.record_feedback(
        {
            "prediction_id": p.prediction_id,
            "category": "accurate",
            "value": abs(p.score - expected_score) < 0.2,
        }
        for expected_score, p in zip(expected_scores, predictions)
    )


def test_record_prediction_feedback_missing_category(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!")
    with pytest.raises(ValueError):
        regression_model.record_feedback({"prediction_id": prediction.prediction_id, "value": True})


def test_record_prediction_feedback_invalid_value(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!")
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        regression_model.record_feedback(
            {"prediction_id": prediction.prediction_id, "category": "accurate", "value": "invalid"}
        )


def test_record_prediction_feedback_invalid_prediction_id(regression_model: RegressionModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        regression_model.record_feedback({"prediction_id": "invalid", "category": "accurate", "value": True})


def test_predict_with_memoryset_override(regression_model: RegressionModel, hf_dataset: Dataset):
    # Create a memoryset with different scores
    inverted_scored_memoryset = ScoredMemoryset.from_hf_dataset(
        "test_memoryset_inverted_scores",
        hf_dataset.map(lambda x: {"score": 2.0 - x["score"]}),  # Invert scores
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
    )
    original_predictions = regression_model.predict(["This is excellent!", "This is terrible!"])

    with regression_model.use_memoryset(inverted_scored_memoryset):
        override_predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
        # With inverted scores, the predictions should be different
        assert abs(override_predictions[0].score - original_predictions[0].score) > 0.1
        assert abs(override_predictions[1].score - original_predictions[1].score) > 0.1

    # After exiting context, predictions should be back to normal
    new_predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    assert abs(new_predictions[0].score - original_predictions[0].score) < 0.1
    assert abs(new_predictions[1].score - original_predictions[1].score) < 0.1


def test_predict_with_expected_scores(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!", expected_scores=0.9)
    assert prediction.expected_score == 0.9


def test_regression_prediction_update(regression_model: RegressionModel):
    prediction = regression_model.predict("Test input", expected_scores=3.5)
    assert prediction.expected_score == 3.5
    assert prediction.tags == set()

    # Update expected score
    prediction.update(expected_score=4.5)
    assert prediction.expected_score == 4.5

    # Add tags
    prediction.update(tags={"test", "updated"})
    assert prediction.tags == {"test", "updated"}

    # Clear both
    prediction.update(expected_score=None, tags=None)
    assert prediction.expected_score is None
    assert prediction.tags == set()


def test_last_prediction_with_batch(regression_model: RegressionModel):
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    assert regression_model.last_prediction is not None
    assert regression_model.last_prediction.prediction_id == predictions[-1].prediction_id
    assert regression_model.last_prediction.input_value == "This is terrible!"
    assert regression_model._last_prediction_was_batch is True


def test_last_prediction_with_single(regression_model: RegressionModel):
    # Test that last_prediction is updated correctly with single prediction
    prediction = regression_model.predict("This is excellent!")
    assert regression_model.last_prediction is not None
    assert regression_model.last_prediction.prediction_id == prediction.prediction_id
    assert regression_model.last_prediction.input_value == "This is excellent!"
    assert regression_model._last_prediction_was_batch is False


def test_batch_predict(regression_model: RegressionModel):
    """Test batch predictions"""
    predictions = regression_model.predict(["test input 1", "test input 2", "test input 3"])
    assert len(predictions) == 3
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)


def test_batch_predict_with_expected_scores(regression_model: RegressionModel):
    """Test batch predictions with expected scores"""
    predictions = regression_model.predict(["input 1", "input 2"], expected_scores=[0.5, 0.8])
    assert len(predictions) == 2
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)


def test_use_memoryset(regression_model: RegressionModel, scored_memoryset: ScoredMemoryset):
    # Test that predictions work with a memoryset
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    assert len(predictions) == 2
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)
    assert all(0 <= pred.confidence <= 1 for pred in predictions)

    # Test that predictions work with a different memoryset
    with regression_model.use_memoryset(scored_memoryset):
        predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
        assert len(predictions) == 2
        assert all(isinstance(pred, RegressionPrediction) for pred in predictions)
        assert all(0 <= pred.confidence <= 1 for pred in predictions)


def test_drop(regression_model):
    """Test that model drop works."""
    name = regression_model.name
    RegressionModel.drop(name)
    assert not RegressionModel.exists(name)



from __future__ import annotations

from typing import Callable, Generic, Iterator, TypedDict, TypeVar, cast, overload

T = TypeVar("T")
R = TypeVar("R")


class Page(TypedDict, Generic[T]):
    items: list[T]
    count: int


class _PagedIterable(Generic[T, R]):
    def __init__(
        self,
        fetch: Callable[[int, int], Page[T]],
        *,
        transform: Callable[[T], R] | None = None,
        page_size: int = 100,
    ) -> None:
        """
        Iterate over a paginated endpoint.

        Parameters:
            fetch: function to fetch a page from the endpoint `(limit: int, offset: int) -> TypedDict[{items: list[T], count: int}]`
            transform: Optional function to transforms item types `(item: T) -> R`, defaults to identity
            limit: maximum number of items to fetch per page
        """
        self.fetch = fetch
        self.transform = transform or (lambda x: cast(R, x))
        self.page_size = page_size
        self.offset = 0  # tracks how much has been yielded, not fetched
        self.page = fetch(self.page_size, self.offset)  # fetch first page to populate count
        self.count = self.page["count"]

    def __iter__(self) -> Iterator[R]:
        if self.offset >= self.count:
            self.offset = 0
            if len(self.page["items"]) < self.count:
                # refetch first page unless we are still on the first page
                self.page = self.fetch(self.page_size, self.offset)

        # yield prefetched first page
        if self.offset == 0:
            yield from map(self.transform, self.page["items"])
            self.offset += len(self.page["items"])

        # yield remaining pages one by one
        while self.offset < self.count:
            self.page = self.fetch(self.page_size, self.offset)
            yield from map(self.transform, self.page["items"])
            self.offset += len(self.page["items"])

    @overload
    def __getitem__(self, key: int) -> R:
        pass

    @overload
    def __getitem__(self, key: slice) -> list[R]:
        pass

    def __getitem__(self, key: int | slice) -> R | list[R]:
        if isinstance(key, int):
            effective_key = key
            if effective_key < 0:
                effective_key += self.count
            if not 0 <= effective_key < self.count:
                raise IndexError(f"Index {key} out of range")
            # if key is on current page, return item
            if self.offset <= effective_key < self.offset + len(self.page["items"]):
                return self.transform(self.page["items"][effective_key - self.offset])
            # otherwise, fetch and return the single item
            return self.transform(self.fetch(1, effective_key)["items"][0])

        elif isinstance(key, slice):
            start, stop, step = key.indices(self.count)
            if step != 1:
                raise ValueError("Stepped slicing is not supported")
            start = start + self.count if start < 0 else start or 0
            stop = stop + self.count if stop < 0 else stop or self.count
            if start >= self.count or stop > self.count:
                raise IndexError(f"Slice {key} out of range")
            limit = min(self.page_size, stop - start)
            if limit <= 0:
                return []
            items = []
            for i in range(start, stop, limit):
                page = self.fetch(limit, i)
                items.extend(map(self.transform, page["items"]))
            return items

    def __len__(self) -> int:
        return self.count


# type checking workaround until python 3.13 allows declaring the class as PagedIterable[T, R = T]


@overload
def PagedIterable(
    fetch: Callable[[int, int], Page[T]],
    *,
    transform: None = None,
    page_size: int = 100,
) -> _PagedIterable[T, T]:
    pass


@overload
def PagedIterable(
    fetch: Callable[[int, int], Page[T]],
    *,
    transform: Callable[[T], R],
    page_size: int = 100,
) -> _PagedIterable[T, R]:
    pass


def PagedIterable(
    fetch: Callable[[int, int], Page[T]],
    *,
    transform: Callable[[T], R] | None = None,
    page_size: int = 100,
) -> _PagedIterable[T, R]:
    return _PagedIterable(fetch, transform=transform, page_size=page_size)



import random
from uuid import uuid4

import pytest
from datasets.arrow_dataset import Dataset

from .classification_model import ClassificationModel
from .conftest import skip_in_prod
from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import LabeledMemoryset, ScoredMemory, ScoredMemoryset, Status

"""
Test Performance Note:

Creating new `LabeledMemoryset` objects is expensive, so this test file applies the following optimizations:

- Two fixtures are used to manage memorysets:
    - `readonly_memoryset` is a session-scoped fixture shared across tests that do not modify state.
      It should only be used in nullipotent tests.
    - `writable_memoryset` is a function-scoped, regenerating fixture.
      It can be used in tests that mutate or delete the memoryset, and will be reset before each test.

- To minimize fixture overhead, tests using `writable_memoryset` should combine related behaviors.
  For example, prefer a single `test_delete` that covers both single and multiple deletion cases,
  rather than separate `test_delete_single` and `test_delete_multiple` tests.
"""


def test_create_memoryset(readonly_memoryset: LabeledMemoryset, hf_dataset: Dataset, label_names: list[str]):
    assert readonly_memoryset is not None
    assert readonly_memoryset.name == "test_readonly_memoryset"
    assert readonly_memoryset.embedding_model == PretrainedEmbeddingModel.GTE_BASE
    assert readonly_memoryset.label_names == label_names
    assert readonly_memoryset.insertion_status == Status.COMPLETED
    assert isinstance(readonly_memoryset.length, int)
    assert readonly_memoryset.length == len(hf_dataset)
    assert readonly_memoryset.index_type == "IVF_FLAT"
    assert readonly_memoryset.index_params == {"n_lists": 100}


def test_create_memoryset_unauthenticated(unauthenticated, datasource):
    with pytest.raises(ValueError, match="Invalid API key"):
        LabeledMemoryset.create("test_memoryset", datasource)


def test_create_memoryset_invalid_input(datasource):
    # invalid name
    with pytest.raises(ValueError, match=r"Invalid input:.*"):
        LabeledMemoryset.create("test memoryset", datasource)


def test_create_memoryset_already_exists_error(hf_dataset, label_names, readonly_memoryset):
    memoryset_name = readonly_memoryset.name
    with pytest.raises(ValueError):
        LabeledMemoryset.from_hf_dataset(memoryset_name, hf_dataset, label_names=label_names)
    with pytest.raises(ValueError):
        LabeledMemoryset.from_hf_dataset(memoryset_name, hf_dataset, label_names=label_names, if_exists="error")


def test_create_memoryset_already_exists_open(hf_dataset, label_names, readonly_memoryset):
    # invalid label names
    with pytest.raises(ValueError):
        LabeledMemoryset.from_hf_dataset(
            readonly_memoryset.name,
            hf_dataset,
            label_names=["turtles", "frogs"],
            if_exists="open",
        )
    # different embedding model
    with pytest.raises(ValueError):
        LabeledMemoryset.from_hf_dataset(
            readonly_memoryset.name,
            hf_dataset,
            label_names=label_names,
            embedding_model=PretrainedEmbeddingModel.DISTILBERT,
            if_exists="open",
        )
    opened_memoryset = LabeledMemoryset.from_hf_dataset(
        readonly_memoryset.name,
        hf_dataset,
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
        if_exists="open",
    )
    assert opened_memoryset is not None
    assert opened_memoryset.name == readonly_memoryset.name
    assert opened_memoryset.length == len(hf_dataset)


def test_open_memoryset(readonly_memoryset, hf_dataset):
    fetched_memoryset = LabeledMemoryset.open(readonly_memoryset.name)
    assert fetched_memoryset is not None
    assert fetched_memoryset.name == readonly_memoryset.name
    assert fetched_memoryset.length == len(hf_dataset)
    assert fetched_memoryset.index_type == "IVF_FLAT"
    assert fetched_memoryset.index_params == {"n_lists": 100}


def test_open_memoryset_unauthenticated(unauthenticated, readonly_memoryset):
    with pytest.raises(ValueError, match="Invalid API key"):
        LabeledMemoryset.open(readonly_memoryset.name)


def test_open_memoryset_not_found():
    with pytest.raises(LookupError):
        LabeledMemoryset.open(str(uuid4()))


def test_open_memoryset_invalid_input():
    with pytest.raises(ValueError, match=r"Invalid input:.*"):
        LabeledMemoryset.open("not valid id")


def test_open_memoryset_unauthorized(unauthorized, readonly_memoryset):
    with pytest.raises(LookupError):
        LabeledMemoryset.open(readonly_memoryset.name)


def test_all_memorysets(readonly_memoryset: LabeledMemoryset):
    memorysets = LabeledMemoryset.all()
    assert len(memorysets) > 0
    assert any(memoryset.name == readonly_memoryset.name for memoryset in memorysets)


def test_all_memorysets_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        LabeledMemoryset.all()


def test_all_memorysets_unauthorized(unauthorized, readonly_memoryset):
    assert readonly_memoryset not in LabeledMemoryset.all()


def test_drop_memoryset_unauthenticated(unauthenticated, readonly_memoryset):
    with pytest.raises(ValueError, match="Invalid API key"):
        LabeledMemoryset.drop(readonly_memoryset.name)


def test_drop_memoryset_not_found():
    with pytest.raises(LookupError):
        LabeledMemoryset.drop(str(uuid4()))
    # ignores error if specified
    LabeledMemoryset.drop(str(uuid4()), if_not_exists="ignore")


def test_drop_memoryset_unauthorized(unauthorized, readonly_memoryset):
    with pytest.raises(LookupError):
        LabeledMemoryset.drop(readonly_memoryset.name)


def test_update_memoryset_attributes(writable_memoryset: LabeledMemoryset):
    original_label_names = writable_memoryset.label_names
    writable_memoryset.set(description="New description")
    assert writable_memoryset.description == "New description"

    writable_memoryset.set(description=None)
    assert writable_memoryset.description is None

    writable_memoryset.set(name="New_name")
    assert writable_memoryset.name == "New_name"

    writable_memoryset.set(name="test_writable_memoryset")
    assert writable_memoryset.name == "test_writable_memoryset"

    assert writable_memoryset.label_names == original_label_names

    writable_memoryset.set(label_names=["New label 1", "New label 2"])
    assert writable_memoryset.label_names == ["New label 1", "New label 2"]


def test_search(readonly_memoryset: LabeledMemoryset):
    memory_lookups = readonly_memoryset.search(["i love soup", "cats are cute"])
    assert len(memory_lookups) == 2
    assert len(memory_lookups[0]) == 1
    assert len(memory_lookups[1]) == 1
    assert memory_lookups[0][0].label == 0
    assert memory_lookups[1][0].label == 1


def test_search_count(readonly_memoryset: LabeledMemoryset):
    memory_lookups = readonly_memoryset.search("i love soup", count=3)
    assert len(memory_lookups) == 3
    assert memory_lookups[0].label == 0
    assert memory_lookups[1].label == 0
    assert memory_lookups[2].label == 0


def test_get_memory_at_index(readonly_memoryset: LabeledMemoryset, hf_dataset: Dataset, label_names: list[str]):
    memory = readonly_memoryset[0]
    assert memory.value == hf_dataset[0]["value"]
    assert memory.label == hf_dataset[0]["label"]
    assert memory.label_name == label_names[hf_dataset[0]["label"]]
    assert memory.source_id == hf_dataset[0]["source_id"]
    assert memory.score == hf_dataset[0]["score"]
    assert memory.key == hf_dataset[0]["key"]
    last_memory = readonly_memoryset[-1]
    assert last_memory.value == hf_dataset[-1]["value"]
    assert last_memory.label == hf_dataset[-1]["label"]


def test_get_range_of_memories(readonly_memoryset: LabeledMemoryset, hf_dataset: Dataset):
    memories = readonly_memoryset[1:3]
    assert len(memories) == 2
    assert memories[0].value == hf_dataset["value"][1]
    assert memories[1].value == hf_dataset["value"][2]


def test_get_memory_by_id(readonly_memoryset: LabeledMemoryset, hf_dataset: Dataset):
    memory = readonly_memoryset.get(readonly_memoryset[0].memory_id)
    assert memory.value == hf_dataset[0]["value"]
    assert memory == readonly_memoryset[memory.memory_id]


def test_get_memories_by_id(readonly_memoryset: LabeledMemoryset, hf_dataset: Dataset):
    memories = readonly_memoryset.get([readonly_memoryset[0].memory_id, readonly_memoryset[1].memory_id])
    assert len(memories) == 2
    assert memories[0].value == hf_dataset[0]["value"]
    assert memories[1].value == hf_dataset[1]["value"]


def test_query_memoryset(readonly_memoryset: LabeledMemoryset):
    memories = readonly_memoryset.query(filters=[("label", "==", 1)])
    assert len(memories) == 8
    assert all(memory.label == 1 for memory in memories)
    assert len(readonly_memoryset.query(limit=2)) == 2
    assert len(readonly_memoryset.query(filters=[("metadata.key", "==", "g2")])) == 4


def test_query_memoryset_with_feedback_metrics(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    feedback_name = f"correct_{random.randint(0, 1000000)}"
    prediction.record_feedback(category=feedback_name, value=prediction.label == 0)
    memories = prediction.memoryset.query(filters=[("label", "==", 0)], with_feedback_metrics=True)

    # Get the memory_ids that were actually used in the prediction
    used_memory_ids = {memory.memory_id for memory in prediction.memory_lookups}

    assert len(memories) == 8
    assert all(memory.label == 0 for memory in memories)
    for memory in memories:
        assert memory.feedback_metrics is not None
        if memory.memory_id in used_memory_ids:
            assert feedback_name in memory.feedback_metrics
            assert memory.feedback_metrics[feedback_name]["avg"] == 1.0
            assert memory.feedback_metrics[feedback_name]["count"] == 1
        else:
            assert feedback_name not in memory.feedback_metrics or memory.feedback_metrics[feedback_name]["count"] == 0
        assert isinstance(memory.lookup_count, int)


def test_query_memoryset_with_feedback_metrics_filter(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    prediction.record_feedback(category="accurate", value=prediction.label == 0)
    memories = prediction.memoryset.query(
        filters=[("feedback_metrics.accurate.avg", ">", 0.5)], with_feedback_metrics=True
    )
    assert len(memories) == 3
    assert all(memory.label == 0 for memory in memories)
    for memory in memories:
        assert memory.feedback_metrics is not None
        assert memory.feedback_metrics["accurate"] is not None
        assert memory.feedback_metrics["accurate"]["avg"] == 1.0
        assert memory.feedback_metrics["accurate"]["count"] == 1


def test_query_memoryset_with_feedback_metrics_sort(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    prediction.record_feedback(category="positive", value=1.0)
    prediction2 = classification_model.predict("Do you like cats?")
    prediction2.record_feedback(category="positive", value=-1.0)

    memories = prediction.memoryset.query(
        filters=[("feedback_metrics.positive.avg", ">=", -1.0)],
        sort=[("feedback_metrics.positive.avg", "desc")],
        with_feedback_metrics=True,
    )
    assert (
        len(memories) == 6
    )  # there are only 6 out of 16 memories that have a positive feedback metric. Look at SAMPLE_DATA in conftest.py
    assert memories[0].feedback_metrics["positive"]["avg"] == 1.0
    assert memories[-1].feedback_metrics["positive"]["avg"] == -1.0


def test_insert_memories(writable_memoryset: LabeledMemoryset):
    writable_memoryset.refresh()
    prev_length = writable_memoryset.length
    writable_memoryset.insert(
        [
            dict(value="tomato soup is my favorite", label=0),
            dict(value="cats are fun to play with", label=1),
        ]
    )
    writable_memoryset.refresh()
    assert writable_memoryset.length == prev_length + 2
    writable_memoryset.insert(dict(value="tomato soup is my favorite", label=0, key="test", source_id="test"))
    writable_memoryset.refresh()
    assert writable_memoryset.length == prev_length + 3
    last_memory = writable_memoryset[-1]
    assert last_memory.value == "tomato soup is my favorite"
    assert last_memory.label == 0
    assert last_memory.metadata
    assert last_memory.metadata["key"] == "test"
    assert last_memory.source_id == "test"


@skip_in_prod("Production memorysets do not have session consistency guarantees")
def test_update_memories(writable_memoryset: LabeledMemoryset, hf_dataset: Dataset):
    # We've combined the update tests into one to avoid multiple expensive requests for a writable_memoryset

    # test updating a single memory
    memory_id = writable_memoryset[0].memory_id
    updated_memory = writable_memoryset.update(dict(memory_id=memory_id, value="i love soup so much"))
    assert updated_memory.value == "i love soup so much"
    assert updated_memory.label == hf_dataset[0]["label"]
    writable_memoryset.refresh()  # Refresh to ensure consistency after update
    assert writable_memoryset.get(memory_id).value == "i love soup so much"

    # test updating a memory instance
    memory = writable_memoryset[0]
    updated_memory = memory.update(value="i love soup even more")
    assert updated_memory is memory
    assert memory.value == "i love soup even more"
    assert memory.label == hf_dataset[0]["label"]

    # test updating multiple memories
    memory_ids = [memory.memory_id for memory in writable_memoryset[:2]]
    updated_memories = writable_memoryset.update(
        [
            dict(memory_id=memory_ids[0], value="i love soup so much"),
            dict(memory_id=memory_ids[1], value="cats are so cute"),
        ]
    )
    assert updated_memories[0].value == "i love soup so much"
    assert updated_memories[1].value == "cats are so cute"


def test_delete_memories(writable_memoryset: LabeledMemoryset):
    # We've combined the delete tests into one to avoid multiple expensive requests for a writable_memoryset

    # test deleting a single memory
    prev_length = writable_memoryset.length
    memory_id = writable_memoryset[0].memory_id
    writable_memoryset.delete(memory_id)
    with pytest.raises(LookupError):
        writable_memoryset.get(memory_id)
    assert writable_memoryset.length == prev_length - 1

    # test deleting multiple memories
    prev_length = writable_memoryset.length
    writable_memoryset.delete([writable_memoryset[0].memory_id, writable_memoryset[1].memory_id])
    assert writable_memoryset.length == prev_length - 2


def test_clone_memoryset(readonly_memoryset: LabeledMemoryset):
    cloned_memoryset = readonly_memoryset.clone(
        "test_cloned_memoryset", embedding_model=PretrainedEmbeddingModel.DISTILBERT
    )
    assert cloned_memoryset is not None
    assert cloned_memoryset.name == "test_cloned_memoryset"
    assert cloned_memoryset.length == readonly_memoryset.length
    assert cloned_memoryset.embedding_model == PretrainedEmbeddingModel.DISTILBERT
    assert cloned_memoryset.insertion_status == Status.COMPLETED


def test_embedding_evaluation(eval_datasource: Datasource):
    results = LabeledMemoryset.run_embedding_evaluation(
        eval_datasource, embedding_models=["CDE_SMALL"], neighbor_count=2
    )
    assert isinstance(results, list)
    assert len(results) == 1
    assert results[0] is not None
    assert results[0]["embedding_model_name"] == "CDE_SMALL"
    assert results[0]["embedding_model_path"] == "OrcaDB/cde-small-v1"


@pytest.fixture(scope="function")
async def test_group_potential_duplicates(writable_memoryset: LabeledMemoryset):
    writable_memoryset.insert(
        [
            dict(value="raspberry soup Is my favorite", label=0),
            dict(value="Raspberry soup is MY favorite", label=0),
            dict(value="rAspberry soup is my favorite", label=0),
            dict(value="raSpberry SOuP is my favorite", label=0),
            dict(value="rasPberry SOuP is my favorite", label=0),
            dict(value="bunny rabbit Is not my mom", label=1),
            dict(value="bunny rabbit is not MY mom", label=1),
            dict(value="bunny rabbit Is not my moM", label=1),
            dict(value="bunny rabbit is not my mom", label=1),
            dict(value="bunny rabbit is not my mom", label=1),
            dict(value="bunny rabbit is not My mom", label=1),
        ]
    )

    writable_memoryset.analyze({"name": "duplicate", "possible_duplicate_threshold": 0.97})
    response = writable_memoryset.get_potential_duplicate_groups()
    assert isinstance(response, list)
    assert sorted([len(res) for res in response]) == [5, 6]  # 5 favorite, 6 mom


def test_get_cascading_edits_suggestions(writable_memoryset: LabeledMemoryset):
    # Insert a memory to test cascading edits
    SOUP = 0
    CATS = 1
    query_text = "i love soup"  # from SAMPLE_DATA in conftest.py
    mislabeled_soup_text = "soup is comfort in a bowl"
    writable_memoryset.insert(
        [
            dict(value=mislabeled_soup_text, label=CATS),  # mislabeled soup memory
        ]
    )

    # Fetch the memory to update
    memory = writable_memoryset.query(filters=[("value", "==", query_text)])[0]

    # Update the label and get cascading edit suggestions
    suggestions = writable_memoryset.get_cascading_edits_suggestions(
        memory=memory,
        old_label=CATS,
        new_label=SOUP,
        max_neighbors=10,
        max_validation_neighbors=5,
    )

    # Validate the suggestions
    assert len(suggestions) == 1
    assert suggestions[0]["neighbor"]["value"] == mislabeled_soup_text


def test_analyze_invalid_analysis_name(readonly_memoryset: LabeledMemoryset):
    """Test that analyze() raises ValueError for invalid analysis names"""
    memoryset = LabeledMemoryset.open(readonly_memoryset.name)

    # Test with string input
    with pytest.raises(ValueError) as excinfo:
        memoryset.analyze("invalid_name")
    assert "Invalid analysis name: invalid_name" in str(excinfo.value)
    assert "Valid names are:" in str(excinfo.value)

    # Test with dict input
    with pytest.raises(ValueError) as excinfo:
        memoryset.analyze({"name": "invalid_name"})
    assert "Invalid analysis name: invalid_name" in str(excinfo.value)
    assert "Valid names are:" in str(excinfo.value)

    # Test with multiple analyses where one is invalid
    with pytest.raises(ValueError) as excinfo:
        memoryset.analyze("duplicate", "invalid_name")
    assert "Invalid analysis name: invalid_name" in str(excinfo.value)
    assert "Valid names are:" in str(excinfo.value)

    # Test with valid analysis names
    result = memoryset.analyze("duplicate", "cluster")
    assert isinstance(result, dict)
    assert "duplicate" in result
    assert "cluster" in result


def test_drop_memoryset(writable_memoryset: LabeledMemoryset):
    # NOTE: Keep this test at the end to ensure the memoryset is dropped after all tests.
    # Otherwise, it would be recreated on the next test run if it were dropped earlier, and
    # that's expensive.
    assert LabeledMemoryset.exists(writable_memoryset.name)
    LabeledMemoryset.drop(writable_memoryset.name)
    assert not LabeledMemoryset.exists(writable_memoryset.name)


def test_scored_memoryset(scored_memoryset: ScoredMemoryset):
    assert scored_memoryset.length == 16
    assert isinstance(scored_memoryset[0], ScoredMemory)
    assert scored_memoryset[0].value == "i love soup"
    assert scored_memoryset[0].score is not None
    assert scored_memoryset[0].metadata == {"key": "g1", "source_id": "s1", "label": 0}
    lookup = scored_memoryset.search("i love soup", count=1)
    assert len(lookup) == 1
    assert lookup[0].score < 0.11


@skip_in_prod("Production memorysets do not have session consistency guarantees")
def test_update_scored_memory(scored_memoryset: ScoredMemoryset):
    # we are only updating an inconsequential metadata field so that we don't affect other tests
    memory = scored_memoryset[0]
    assert memory.label == 0
    scored_memoryset.update(dict(memory_id=memory.memory_id, label=3))
    assert scored_memoryset[0].label == 3
    memory.update(label=4)
    assert scored_memoryset[0].label == 4



"""This module contains internal utils for managing api keys in tests"""

import logging
import os
from typing import List, Literal

from dotenv import load_dotenv

from ..client import ApiKeyMetadata, orca_api
from ..credentials import OrcaCredentials
from .common import DropMode

load_dotenv()  # this needs to be here to ensure env is populated before accessing it

# the defaults here must match nautilus and lighthouse config defaults
_ORCA_ROOT_ACCESS_API_KEY = os.environ.get("ORCA_ROOT_ACCESS_API_KEY", "00000000-0000-0000-0000-000000000000")
_DEFAULT_ORG_ID = os.environ.get("DEFAULT_ORG_ID", "10e50000-0000-4000-a000-a78dca14af3a")


def _create_api_key(org_id: str, name: str, scopes: list[Literal["ADMINISTER", "PREDICT"]] = ["ADMINISTER"]) -> str:
    """Creates an API key for the given organization"""
    response = orca_api.POST(
        "/auth/api_key",
        json={"name": name, "scope": scopes},
        headers={"Api-Key": _ORCA_ROOT_ACCESS_API_KEY, "Org-Id": org_id},
    )
    return response["api_key"]


def _list_api_keys(org_id: str) -> List[ApiKeyMetadata]:
    """Lists all API keys for the given organization"""
    return orca_api.GET("/auth/api_key", headers={"Api-Key": _ORCA_ROOT_ACCESS_API_KEY, "Org-Id": org_id})


def _delete_api_key(org_id: str, name: str, if_not_exists: DropMode = "error") -> None:
    """Deletes the API key with the given name from the organization"""
    try:
        orca_api.DELETE(
            "/auth/api_key/{name_or_id}",
            params={"name_or_id": name},
            headers={"Api-Key": _ORCA_ROOT_ACCESS_API_KEY, "Org-Id": org_id},
        )
    except LookupError:
        if if_not_exists == "error":
            raise


def _delete_org(org_id: str) -> None:
    """Deletes the organization"""
    orca_api.DELETE("/auth/org", headers={"Api-Key": _ORCA_ROOT_ACCESS_API_KEY, "Org-Id": org_id})


def _authenticate_local_api(org_id: str = _DEFAULT_ORG_ID, api_key_name: str = "local") -> None:
    """Connect to the local API at http://localhost:1584/ and authenticate with a new API key"""
    _delete_api_key(org_id, api_key_name, if_not_exists="ignore")
    OrcaCredentials.set_base_url("http://localhost:1584")
    OrcaCredentials.set_api_key(_create_api_key(org_id, api_key_name))
    logging.info(f"Authenticated against local API at 'http://localhost:1584' with '{api_key_name}' API key")


__all__ = ["_create_api_key", "_delete_api_key", "_delete_org", "_list_api_keys", "_authenticate_local_api"]



"""
IMPORTANT:
- This is a shared file between OrcaLib and the Orca SDK.
- Please ensure that it does not have any dependencies on the OrcaLib code.
- Make sure to edit this file in orcalib/shared and NOT in orca_sdk, since it will be overwritten there.
"""

from typing import Literal

import numpy as np
import pytest
import sklearn.metrics

from .metrics import (
    calculate_classification_metrics,
    calculate_pr_curve,
    calculate_regression_metrics,
    calculate_roc_curve,
    softmax,
)


def test_binary_metrics():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([0.1, 0.9, 0.8, 0.3, 0.2])

    metrics = calculate_classification_metrics(y_true, y_score)

    assert metrics.accuracy == 0.8
    assert metrics.f1_score == 0.8
    assert metrics.roc_auc is not None
    assert metrics.roc_auc > 0.8
    assert metrics.roc_auc < 1.0
    assert metrics.pr_auc is not None
    assert metrics.pr_auc > 0.8
    assert metrics.pr_auc < 1.0
    assert metrics.loss is not None
    assert metrics.loss > 0.0


def test_multiclass_metrics_with_2_classes():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([[0.9, 0.1], [0.1, 0.9], [0.2, 0.8], [0.7, 0.3], [0.8, 0.2]])

    metrics = calculate_classification_metrics(y_true, y_score)

    assert metrics.accuracy == 0.8
    assert metrics.f1_score == 0.8
    assert metrics.roc_auc is not None
    assert metrics.roc_auc > 0.8
    assert metrics.roc_auc < 1.0
    assert metrics.pr_auc is not None
    assert metrics.pr_auc > 0.8
    assert metrics.pr_auc < 1.0
    assert metrics.loss is not None
    assert metrics.loss > 0.0


@pytest.mark.parametrize(
    "average, multiclass",
    [("micro", "ovr"), ("macro", "ovr"), ("weighted", "ovr"), ("micro", "ovo"), ("macro", "ovo"), ("weighted", "ovo")],
)
def test_multiclass_metrics_with_3_classes(
    average: Literal["micro", "macro", "weighted"], multiclass: Literal["ovr", "ovo"]
):
    y_true = np.array([0, 1, 1, 0, 2])
    y_score = np.array([[0.9, 0.1, 0.0], [0.1, 0.9, 0.0], [0.2, 0.8, 0.0], [0.7, 0.3, 0.0], [0.0, 0.0, 1.0]])

    metrics = calculate_classification_metrics(y_true, y_score, average=average, multi_class=multiclass)

    assert metrics.accuracy == 1.0
    assert metrics.f1_score == 1.0
    assert metrics.roc_auc is not None
    assert metrics.roc_auc > 0.8
    assert metrics.pr_auc is None
    assert metrics.loss is not None
    assert metrics.loss > 0.0


def test_does_not_modify_logits_unless_necessary():
    logits = np.array([[0.1, 0.9], [0.2, 0.8], [0.7, 0.3], [0.8, 0.2]])
    expected_labels = [0, 1, 0, 1]
    assert calculate_classification_metrics(expected_labels, logits).loss == sklearn.metrics.log_loss(
        expected_labels, logits
    )


def test_normalizes_logits_if_necessary():
    logits = np.array([[1.2, 3.9], [1.2, 5.8], [1.2, 2.7], [1.2, 1.3]])
    expected_labels = [0, 1, 0, 1]
    assert calculate_classification_metrics(expected_labels, logits).loss == sklearn.metrics.log_loss(
        expected_labels, logits / logits.sum(axis=1, keepdims=True)
    )


def test_softmaxes_logits_if_necessary():
    logits = np.array([[-1.2, 3.9], [1.2, -5.8], [1.2, 2.7], [1.2, 1.3]])
    expected_labels = [0, 1, 0, 1]
    assert calculate_classification_metrics(expected_labels, logits).loss == sklearn.metrics.log_loss(
        expected_labels, softmax(logits)
    )


def test_precision_recall_curve():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([0.1, 0.9, 0.8, 0.6, 0.2])

    pr_curve = calculate_pr_curve(y_true, y_score)

    assert len(pr_curve["precisions"]) == len(pr_curve["recalls"]) == len(pr_curve["thresholds"]) == 6
    assert np.allclose(pr_curve["precisions"][0], 0.6)
    assert np.allclose(pr_curve["recalls"][0], 1.0)
    assert np.allclose(pr_curve["precisions"][-1], 1.0)
    assert np.allclose(pr_curve["recalls"][-1], 0.0)

    # test that thresholds are sorted
    assert np.all(np.diff(pr_curve["thresholds"]) >= 0)


def test_roc_curve():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([0.1, 0.9, 0.8, 0.6, 0.2])

    roc_curve = calculate_roc_curve(y_true, y_score)

    assert (
        len(roc_curve["false_positive_rates"])
        == len(roc_curve["true_positive_rates"])
        == len(roc_curve["thresholds"])
        == 6
    )
    assert roc_curve["false_positive_rates"][0] == 1.0
    assert roc_curve["true_positive_rates"][0] == 1.0
    assert roc_curve["false_positive_rates"][-1] == 0.0
    assert roc_curve["true_positive_rates"][-1] == 0.0

    # test that thresholds are sorted
    assert np.all(np.diff(roc_curve["thresholds"]) >= 0)


def test_precision_recall_curve_max_length():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([0.1, 0.9, 0.8, 0.6, 0.2])

    pr_curve = calculate_pr_curve(y_true, y_score, max_length=5)
    assert len(pr_curve["precisions"]) == len(pr_curve["recalls"]) == len(pr_curve["thresholds"]) == 5

    assert np.allclose(pr_curve["precisions"][0], 0.6)
    assert np.allclose(pr_curve["recalls"][0], 1.0)
    assert np.allclose(pr_curve["precisions"][-1], 1.0)
    assert np.allclose(pr_curve["recalls"][-1], 0.0)

    # test that thresholds are sorted
    assert np.all(np.diff(pr_curve["thresholds"]) >= 0)


def test_roc_curve_max_length():
    y_true = np.array([0, 1, 1, 0, 1])
    y_score = np.array([0.1, 0.9, 0.8, 0.6, 0.2])

    roc_curve = calculate_roc_curve(y_true, y_score, max_length=5)
    assert (
        len(roc_curve["false_positive_rates"])
        == len(roc_curve["true_positive_rates"])
        == len(roc_curve["thresholds"])
        == 5
    )
    assert np.allclose(roc_curve["false_positive_rates"][0], 1.0)
    assert np.allclose(roc_curve["true_positive_rates"][0], 1.0)
    assert np.allclose(roc_curve["false_positive_rates"][-1], 0.0)
    assert np.allclose(roc_curve["true_positive_rates"][-1], 0.0)

    # test that thresholds are sorted
    assert np.all(np.diff(roc_curve["thresholds"]) >= 0)


# Regression Metrics Tests


def test_perfect_regression_predictions():
    y_true = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
    y_pred = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)

    metrics = calculate_regression_metrics(y_true, y_pred)

    assert metrics.mse == 0.0
    assert metrics.rmse == 0.0
    assert metrics.mae == 0.0
    assert metrics.r2 == 1.0
    assert metrics.explained_variance == 1.0
    assert metrics.loss == 0.0
    assert metrics.anomaly_score_mean is None
    assert metrics.anomaly_score_median is None
    assert metrics.anomaly_score_variance is None


def test_basic_regression_metrics():
    y_true = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
    y_pred = np.array([1.1, 1.9, 3.2, 3.8, 5.1], dtype=np.float32)

    metrics = calculate_regression_metrics(y_true, y_pred)

    # Check that all metrics are reasonable
    assert metrics.mse > 0.0
    assert metrics.rmse == pytest.approx(np.sqrt(metrics.mse))
    assert metrics.mae > 0.0
    assert 0.0 <= metrics.r2 <= 1.0
    assert 0.0 <= metrics.explained_variance <= 1.0
    assert metrics.loss == metrics.mse

    # Check specific values based on the data
    expected_mse = np.mean((y_true - y_pred) ** 2)
    assert metrics.mse == pytest.approx(expected_mse)

    expected_mae = np.mean(np.abs(y_true - y_pred))
    assert metrics.mae == pytest.approx(expected_mae)


def test_regression_metrics_with_anomaly_scores():
    y_true = np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
    y_pred = np.array([1.1, 1.9, 3.2, 3.8, 5.1], dtype=np.float32)
    anomaly_scores = [0.1, 0.2, 0.15, 0.3, 0.25]

    metrics = calculate_regression_metrics(y_true, y_pred, anomaly_scores)

    assert metrics.anomaly_score_mean == pytest.approx(np.mean(anomaly_scores))
    assert metrics.anomaly_score_median == pytest.approx(np.median(anomaly_scores))
    assert metrics.anomaly_score_variance == pytest.approx(np.var(anomaly_scores))



import logging
from uuid import uuid4

import numpy as np
import pytest
from datasets.arrow_dataset import Dataset

from .classification_model import ClassificationMetrics, ClassificationModel
from .conftest import skip_in_ci
from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import LabeledMemoryset


def test_create_model(classification_model: ClassificationModel, readonly_memoryset: LabeledMemoryset):
    assert classification_model is not None
    assert classification_model.name == "test_classification_model"
    assert classification_model.memoryset == readonly_memoryset
    assert classification_model.num_classes == 2
    assert classification_model.memory_lookup_count == 3


def test_create_model_already_exists_error(readonly_memoryset, classification_model):
    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset)
    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="error")


def test_create_model_already_exists_return(readonly_memoryset, classification_model):
    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="open", head_type="MMOE")

    with pytest.raises(ValueError):
        ClassificationModel.create(
            "test_classification_model", readonly_memoryset, if_exists="open", memory_lookup_count=37
        )

    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="open", num_classes=19)

    with pytest.raises(ValueError):
        ClassificationModel.create(
            "test_classification_model", readonly_memoryset, if_exists="open", min_memory_weight=0.77
        )

    new_model = ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_classification_model"
    assert new_model.memoryset == readonly_memoryset
    assert new_model.num_classes == 2
    assert new_model.memory_lookup_count == 3


def test_create_model_unauthenticated(unauthenticated, readonly_memoryset: LabeledMemoryset):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.create("test_model", readonly_memoryset)


def test_get_model(classification_model: ClassificationModel):
    fetched_model = ClassificationModel.open(classification_model.name)
    assert fetched_model is not None
    assert fetched_model.id == classification_model.id
    assert fetched_model.name == classification_model.name
    assert fetched_model.num_classes == 2
    assert fetched_model.memory_lookup_count == 3
    assert fetched_model == classification_model


def test_get_model_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.open("test_model")


def test_get_model_invalid_input():
    with pytest.raises(ValueError, match="Invalid input"):
        ClassificationModel.open("not valid id")


def test_get_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.open(str(uuid4()))


def test_get_model_unauthorized(unauthorized, classification_model: ClassificationModel):
    with pytest.raises(LookupError):
        ClassificationModel.open(classification_model.name)


def test_list_models(classification_model: ClassificationModel):
    models = ClassificationModel.all()
    assert len(models) > 0
    assert any(model.name == model.name for model in models)


def test_list_models_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.all()


def test_list_models_unauthorized(unauthorized, classification_model: ClassificationModel):
    assert ClassificationModel.all() == []


def test_update_model_attributes(classification_model: ClassificationModel):
    classification_model.description = "New description"
    assert classification_model.description == "New description"

    classification_model.set(description=None)
    assert classification_model.description is None

    classification_model.set(locked=True)
    assert classification_model.locked is True

    classification_model.set(locked=False)
    assert classification_model.locked is False

    classification_model.lock()
    assert classification_model.locked is True

    classification_model.unlock()
    assert classification_model.locked is False


def test_delete_model(readonly_memoryset: LabeledMemoryset):
    ClassificationModel.create("model_to_delete", LabeledMemoryset.open(readonly_memoryset.name))
    assert ClassificationModel.open("model_to_delete")
    ClassificationModel.drop("model_to_delete")
    with pytest.raises(LookupError):
        ClassificationModel.open("model_to_delete")


def test_delete_model_unauthenticated(unauthenticated, classification_model: ClassificationModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.drop(classification_model.name)


def test_delete_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.drop(str(uuid4()))
    # ignores error if specified
    ClassificationModel.drop(str(uuid4()), if_not_exists="ignore")


def test_delete_model_unauthorized(unauthorized, classification_model: ClassificationModel):
    with pytest.raises(LookupError):
        ClassificationModel.drop(classification_model.name)


def test_delete_memoryset_before_model_constraint_violation(hf_dataset):
    memoryset = LabeledMemoryset.from_hf_dataset("test_memoryset_delete_before_model", hf_dataset)
    ClassificationModel.create("test_model_delete_before_memoryset", memoryset)
    with pytest.raises(RuntimeError):
        LabeledMemoryset.drop(memoryset.id)


@pytest.mark.parametrize("data_type", ["dataset", "datasource"])
def test_evaluate(classification_model, eval_datasource: Datasource, eval_dataset: Dataset, data_type):
    result = (
        classification_model.evaluate(eval_dataset)
        if data_type == "dataset"
        else classification_model.evaluate(eval_datasource)
    )

    assert result is not None
    assert isinstance(result, ClassificationMetrics)

    assert isinstance(result.accuracy, float)
    assert np.allclose(result.accuracy, 0.5)
    assert isinstance(result.f1_score, float)
    assert np.allclose(result.f1_score, 0.5)
    assert isinstance(result.loss, float)

    assert isinstance(result.anomaly_score_mean, float)
    assert isinstance(result.anomaly_score_median, float)
    assert isinstance(result.anomaly_score_variance, float)
    assert -1.0 <= result.anomaly_score_mean <= 1.0
    assert -1.0 <= result.anomaly_score_median <= 1.0
    assert -1.0 <= result.anomaly_score_variance <= 1.0

    assert result.pr_auc is not None
    assert np.allclose(result.pr_auc, 0.75)
    assert result.pr_curve is not None
    assert np.allclose(result.pr_curve["thresholds"], [0.0, 0.0, 0.8155114054679871, 0.834095299243927])
    assert np.allclose(result.pr_curve["precisions"], [0.5, 0.5, 1.0, 1.0])
    assert np.allclose(result.pr_curve["recalls"], [1.0, 0.5, 0.5, 0.0])

    assert result.roc_auc is not None
    assert np.allclose(result.roc_auc, 0.625)
    assert result.roc_curve is not None
    assert np.allclose(result.roc_curve["thresholds"], [0.0, 0.8155114054679871, 0.834095299243927, 1.0])
    assert np.allclose(result.roc_curve["false_positive_rates"], [1.0, 0.5, 0.0, 0.0])
    assert np.allclose(result.roc_curve["true_positive_rates"], [1.0, 0.5, 0.5, 0.0])


def test_evaluate_with_telemetry(classification_model: ClassificationModel, eval_dataset: Dataset):
    result = classification_model.evaluate(eval_dataset, record_predictions=True, tags={"test"})
    assert result is not None
    assert isinstance(result, ClassificationMetrics)
    predictions = classification_model.predictions(tag="test")
    assert len(predictions) == 4
    assert all(p.tags == {"test"} for p in predictions)
    assert all(p.expected_label == l for p, l in zip(predictions, eval_dataset["label"]))


def test_predict(classification_model: ClassificationModel, label_names: list[str]):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1

    assert predictions[0].logits is not None
    assert predictions[1].logits is not None
    assert len(predictions[0].logits) == 2
    assert len(predictions[1].logits) == 2
    assert predictions[0].logits[0] > predictions[0].logits[1]
    assert predictions[1].logits[0] < predictions[1].logits[1]


def test_predict_disable_telemetry(classification_model: ClassificationModel, label_names: list[str]):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"], save_telemetry="off")
    assert len(predictions) == 2
    assert predictions[0].prediction_id is None
    assert predictions[1].prediction_id is None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1


def test_predict_unauthenticated(unauthenticated, classification_model: ClassificationModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        classification_model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_unauthorized(unauthorized, classification_model: ClassificationModel):
    with pytest.raises(LookupError):
        classification_model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_constraint_violation(readonly_memoryset: LabeledMemoryset):
    model = ClassificationModel.create(
        "test_model_lookup_count_too_high",
        readonly_memoryset,
        num_classes=2,
        memory_lookup_count=readonly_memoryset.length + 2,
    )
    with pytest.raises(RuntimeError):
        model.predict("test")


def test_record_prediction_feedback(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    expected_labels = [0, 1]
    classification_model.record_feedback(
        {
            "prediction_id": p.prediction_id,
            "category": "correct",
            "value": p.label == expected_label,
        }
        for expected_label, p in zip(expected_labels, predictions)
    )


def test_record_prediction_feedback_missing_category(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    with pytest.raises(ValueError):
        classification_model.record_feedback({"prediction_id": prediction.prediction_id, "value": True})


def test_record_prediction_feedback_invalid_value(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.record_feedback(
            {"prediction_id": prediction.prediction_id, "category": "correct", "value": "invalid"}
        )


def test_record_prediction_feedback_invalid_prediction_id(classification_model: ClassificationModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.record_feedback({"prediction_id": "invalid", "category": "correct", "value": True})


def test_predict_with_memoryset_override(classification_model: ClassificationModel, hf_dataset: Dataset):
    inverted_labeled_memoryset = LabeledMemoryset.from_hf_dataset(
        "test_memoryset_inverted_labels",
        hf_dataset.map(lambda x: {"label": 1 if x["label"] == 0 else 0}),
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
    )
    with classification_model.use_memoryset(inverted_labeled_memoryset):
        predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
        assert predictions[0].label == 1
        assert predictions[1].label == 0

    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert predictions[0].label == 0
    assert predictions[1].label == 1


def test_predict_with_expected_labels(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?", expected_labels=1)
    assert prediction.expected_label == 1


def test_predict_with_expected_labels_invalid_input(classification_model: ClassificationModel):
    # invalid number of expected labels for batch prediction
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.predict(["Do you love soup?", "Are cats cute?"], expected_labels=[0])
    # invalid label value
    with pytest.raises(ValueError):
        classification_model.predict("Do you love soup?", expected_labels=5)


def test_predict_with_filters(classification_model: ClassificationModel):
    # there are no memories with label 0 and key g1, so we force a wrong prediction
    filtered_prediction = classification_model.predict("I love soup", filters=[("key", "==", "g2")])
    assert filtered_prediction.label == 1
    assert filtered_prediction.label_name == "cats"


def test_predict_with_memoryset_update(writable_memoryset: LabeledMemoryset):
    model = ClassificationModel.create(
        "test_predict_with_memoryset_update",
        writable_memoryset,
        num_classes=2,
        memory_lookup_count=3,
    )

    prediction = model.predict("Do you love soup?")
    assert prediction.label == 0
    assert prediction.label_name == "soup"

    # insert new memories
    writable_memoryset.insert(
        [
            {"value": "Do you love soup?", "label": 1, "key": "g1"},
            {"value": "Do you love soup for dinner?", "label": 1, "key": "g2"},
            {"value": "Do you love crackers?", "label": 1, "key": "g2"},
            {"value": "Do you love broth?", "label": 1, "key": "g2"},
            {"value": "Do you love chicken soup?", "label": 1, "key": "g2"},
            {"value": "Do you love chicken soup for dinner?", "label": 1, "key": "g2"},
            {"value": "Do you love chicken soup for dinner?", "label": 1, "key": "g2"},
        ],
    )
    prediction = model.predict("Do you love soup?")
    assert prediction.label == 1
    assert prediction.label_name == "cats"

    ClassificationModel.drop("test_predict_with_memoryset_update")


def test_last_prediction_with_batch(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert classification_model.last_prediction is not None
    assert classification_model.last_prediction.prediction_id == predictions[-1].prediction_id
    assert classification_model.last_prediction.input_value == "Are cats cute?"
    assert classification_model._last_prediction_was_batch is True


def test_last_prediction_with_single(classification_model: ClassificationModel):
    # Test that last_prediction is updated correctly with single prediction
    prediction = classification_model.predict("Do you love soup?")
    assert classification_model.last_prediction is not None
    assert classification_model.last_prediction.prediction_id == prediction.prediction_id
    assert classification_model.last_prediction.input_value == "Do you love soup?"
    assert classification_model._last_prediction_was_batch is False


@skip_in_ci("We don't have Anthropic API key in CI")
def test_explain(writable_memoryset: LabeledMemoryset):

    writable_memoryset.analyze(
        {"name": "neighbor", "neighbor_counts": [1, 3]},
        lookup_count=3,
    )

    model = ClassificationModel.create(
        "test_model_for_explain",
        writable_memoryset,
        num_classes=2,
        memory_lookup_count=3,
        description="This is a test model for explain",
    )

    predictions = model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2

    try:
        explanation = predictions[0].explanation
        assert explanation is not None
        assert len(explanation) > 10
        assert "soup" in explanation.lower()
    except Exception as e:
        if "ANTHROPIC_API_KEY" in str(e):
            logging.info("Skipping explanation test because ANTHROPIC_API_KEY is not set")
        else:
            raise e
    finally:
        ClassificationModel.drop("test_model_for_explain")


@skip_in_ci("We don't have Anthropic API key in CI")
def test_action_recommendation(writable_memoryset: LabeledMemoryset):
    """Test getting action recommendations for predictions"""

    writable_memoryset.analyze(
        {"name": "neighbor", "neighbor_counts": [1, 3]},
        lookup_count=3,
    )

    model = ClassificationModel.create(
        "test_model_for_action",
        writable_memoryset,
        num_classes=2,
        memory_lookup_count=3,
        description="This is a test model for action recommendations",
    )

    # Make a prediction with expected label to simulate incorrect prediction
    prediction = model.predict("Do you love soup?", expected_labels=1)

    memoryset_length = model.memoryset.length

    try:
        # Get action recommendation
        action, rationale = prediction.recommend_action()

        assert action is not None
        assert rationale is not None
        assert action in ["remove_duplicates", "detect_mislabels", "add_memories", "finetuning"]
        assert len(rationale) > 10

        # Test memory suggestions
        suggestions_response = prediction.generate_memory_suggestions(num_memories=2)
        memory_suggestions = suggestions_response.suggestions

        assert memory_suggestions is not None
        assert len(memory_suggestions) == 2

        for suggestion in memory_suggestions:
            assert isinstance(suggestion[0], str)
            assert len(suggestion[0]) > 0
            assert isinstance(suggestion[1], str)
            assert suggestion[1] in model.memoryset.label_names

        suggestions_response.apply()

        model.memoryset.refresh()
        assert model.memoryset.length == memoryset_length + 2

    except Exception as e:
        if "ANTHROPIC_API_KEY" in str(e):
            logging.info("Skipping agent tests because ANTHROPIC_API_KEY is not set")
        else:
            raise e
    finally:
        ClassificationModel.drop("test_model_for_action")


def test_predict_with_prompt(classification_model: ClassificationModel):
    """Test that prompt parameter is properly passed through to predictions"""
    # Test with an instruction-supporting embedding model if available
    prediction_with_prompt = classification_model.predict(
        "I love this product!", prompt="Represent this text for sentiment classification:"
    )
    prediction_without_prompt = classification_model.predict("I love this product!")

    # Both should work and return valid predictions
    assert prediction_with_prompt.label is not None
    assert prediction_without_prompt.label is not None



from datetime import datetime
from typing import Literal, NamedTuple

import httpx
from httpx import ConnectError, Headers

from .client import orca_api

Scope = Literal["ADMINISTER", "PREDICT"]
"""
The scopes of an API key.

- `ADMINISTER`: Can do anything, including creating and deleting organizations, models, and API keys.
- `PREDICT`: Can only call model.predict and perform CRUD operations on predictions.
"""


class ApiKeyInfo(NamedTuple):
    """
    Named tuple containing information about an API key

    Attributes:
        name: Unique name of the API key
        created_at: When the API key was created
    """

    name: str
    created_at: datetime
    scopes: set[Scope]


class OrcaCredentials:
    """
    Class for managing Orca API credentials
    """

    @staticmethod
    def get_api_url() -> str:
        """
        Get the Orca API base URL that is currently being used
        """
        return str(orca_api.base_url)

    @staticmethod
    def list_api_keys() -> list[ApiKeyInfo]:
        """
        List all API keys that have been created for your org

        Returns:
            A list of named tuples, with the name and creation date time of the API key
        """
        return [
            ApiKeyInfo(
                name=api_key["name"],
                created_at=datetime.fromisoformat(api_key["created_at"]),
                scopes=set(api_key["scope"]),
            )
            for api_key in orca_api.GET("/auth/api_key")
        ]

    @staticmethod
    def is_authenticated() -> bool:
        """
        Check if you are authenticated to interact with the Orca API

        Returns:
            True if you are authenticated, False otherwise
        """
        try:
            return orca_api.GET("/auth")
        except ValueError as e:
            if "Invalid API key" in str(e):
                return False
            raise e

    @staticmethod
    def create_api_key(name: str, scopes: set[Scope] = {"ADMINISTER"}) -> str:
        """
        Create a new API key with the given name and scopes

        Params:
            name: The name of the API key
            scopes: The scopes of the API key

        Returns:
            The secret value of the API key. Make sure to save this value as it will not be shown again.
        """
        res = orca_api.POST(
            "/auth/api_key",
            json={"name": name, "scope": list(scopes)},
        )
        return res["api_key"]

    @staticmethod
    def revoke_api_key(name: str) -> None:
        """
        Delete an API key

        Params:
            name: The name of the API key to delete

        Raises:
            ValueError: if the API key is not found
        """
        orca_api.DELETE("/auth/api_key/{name_or_id}", params={"name_or_id": name})

    @staticmethod
    def set_headers(headers: dict[str, str]):
        """
        Add or override default HTTP headers for all Orca API requests.

        Args:
            **kwargs: Header names with their string values

        Notes:
            New keys are merged into the existing headers, this will overwrite headers with the
            same name, but leave other headers untouched.
        """
        orca_api.headers.update(Headers(headers))

    @staticmethod
    def set_api_key(api_key: str, check_validity: bool = True):
        """
        Set the API key to use for authenticating with the Orca API

        Note:
            The API key can also be provided by setting the `ORCA_API_KEY` environment variable

        Params:
            api_key: The API key to set
            check_validity: Whether to check if the API key is valid and raise an error otherwise

        Raises:
            ValueError: if the API key is invalid and `check_validity` is True
        """
        OrcaCredentials.set_headers({"Api-Key": api_key})
        if check_validity:
            orca_api.GET("/auth")

    @staticmethod
    def set_base_url(base_url: str, check_validity: bool = True):
        """
        Set the base URL for the Orca API

        Args:
            base_url: The base URL to set
            check_validity: Whether to check if there is an API running at the given base URL

        Raises:
            ValueError: if there is no healthy API running at the given base URL and `check_validity` is True
        """
        # check if the base url is reachable before setting it
        if check_validity:
            try:
                httpx.get(base_url, timeout=1)
            except ConnectError as e:
                raise ValueError(f"No API found at {base_url}") from e

        orca_api.base_url = base_url

        # check if the api passes the health check
        if check_validity:
            orca_api.GET("/")

    @staticmethod
    def is_healthy() -> bool:
        """
        Check whether the API is healthy

        Returns:
            True if the API is healthy, False otherwise
        """
        try:
            orca_api.GET("/")
            orca_api.GET("/gpu/")
        except Exception:
            return False
        return True



from __future__ import annotations

import logging
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Generator, Iterable, Literal, cast, overload

from datasets import Dataset

from ._shared.metrics import RegressionMetrics, calculate_regression_metrics
from ._utils.common import UNSET, CreateMode, DropMode
from .client import (
    PredictiveModelUpdate,
    RARHeadType,
    RegressionModelMetadata,
    orca_api,
)
from .datasource import Datasource
from .job import Job
from .memoryset import ScoredMemoryset
from .telemetry import (
    RegressionPrediction,
    TelemetryMode,
    _get_telemetry_config,
    _parse_feedback,
)

logger = logging.getLogger(__name__)


class RegressionModel:
    """
    A handle to a regression model in OrcaCloud

    Attributes:
        id: Unique identifier for the model
        name: Unique name of the model
        description: Optional description of the model
        memoryset: Memoryset that the model uses
        head_type: Regression head type of the model
        memory_lookup_count: Number of memories the model uses for each prediction
        locked: Whether the model is locked to prevent accidental deletion
        created_at: When the model was created
        updated_at: When the model was last updated
    """

    id: str
    name: str
    description: str | None
    memoryset: ScoredMemoryset
    head_type: RARHeadType
    memory_lookup_count: int
    version: int
    locked: bool
    created_at: datetime
    updated_at: datetime
    memoryset_id: str

    _last_prediction: RegressionPrediction | None
    _last_prediction_was_batch: bool
    _memoryset_override_id: str | None

    def __init__(self, metadata: RegressionModelMetadata):
        # for internal use only, do not document
        self.id = metadata["id"]
        self.name = metadata["name"]
        self.description = metadata["description"]
        self.memoryset = ScoredMemoryset.open(metadata["memoryset_id"])
        self.head_type = metadata["head_type"]
        self.memory_lookup_count = metadata["memory_lookup_count"]
        self.version = metadata["version"]
        self.locked = metadata["locked"]
        self.created_at = datetime.fromisoformat(metadata["created_at"])
        self.updated_at = datetime.fromisoformat(metadata["updated_at"])
        self.memoryset_id = metadata["memoryset_id"]

        self._memoryset_override_id = None
        self._last_prediction = None
        self._last_prediction_was_batch = False

    def __eq__(self, other) -> bool:
        return isinstance(other, RegressionModel) and self.id == other.id

    def __repr__(self):
        return (
            "RegressionModel({\n"
            f"    name: '{self.name}',\n"
            f"    head_type: {self.head_type},\n"
            f"    memory_lookup_count: {self.memory_lookup_count},\n"
            f"    memoryset: ScoredMemoryset.open('{self.memoryset.name}'),\n"
            "})"
        )

    @property
    def last_prediction(self) -> RegressionPrediction:
        """
        Last prediction made by the model

        Note:
            If the last prediction was part of a batch prediction, the last prediction from the
            batch is returned. If no prediction has been made yet, a [`LookupError`][LookupError]
            is raised.
        """
        if self._last_prediction_was_batch:
            logging.warning(
                "Last prediction was part of a batch prediction, returning the last prediction from the batch"
            )
        if self._last_prediction is None:
            raise LookupError("No prediction has been made yet")
        return self._last_prediction

    @classmethod
    def create(
        cls,
        name: str,
        memoryset: ScoredMemoryset,
        memory_lookup_count: int | None = None,
        description: str | None = None,
        if_exists: CreateMode = "error",
    ) -> RegressionModel:
        """
        Create a regression model.

        Params:
            name: Name of the model
            memoryset: The scored memoryset to use for prediction
            memory_lookup_count: Number of memories to retrieve for prediction. Defaults to 10.
            description: Description of the model
            if_exists: How to handle existing models with the same name

        Returns:
            RegressionModel instance

        Raises:
            ValueError: If a model with the same name already exists and if_exists is "error"
            ValueError: If the memoryset is empty
            ValueError: If memory_lookup_count exceeds the number of memories in the memoryset
        """
        existing = cls.exists(name)
        if existing:
            if if_exists == "error":
                raise ValueError(f"RegressionModel with name '{name}' already exists")
            elif if_exists == "open":
                existing = cls.open(name)
                for attribute in {"memory_lookup_count"}:
                    local_attribute = locals()[attribute]
                    existing_attribute = getattr(existing, attribute)
                    if local_attribute is not None and local_attribute != existing_attribute:
                        raise ValueError(f"Model with name {name} already exists with different {attribute}")

                # special case for memoryset
                if existing.memoryset_id != memoryset.id:
                    raise ValueError(f"Model with name {name} already exists with different memoryset")

                return existing

        metadata = orca_api.POST(
            "/regression_model",
            json={
                "name": name,
                "memoryset_name_or_id": memoryset.id,
                "memory_lookup_count": memory_lookup_count,
                "description": description,
            },
        )
        return cls(metadata)

    @classmethod
    def open(cls, name: str) -> RegressionModel:
        """
        Get a handle to a regression model in the OrcaCloud

        Params:
            name: Name or unique identifier of the regression model

        Returns:
            Handle to the existing regression model in the OrcaCloud

        Raises:
            LookupError: If the regression model does not exist
        """
        return cls(orca_api.GET("/regression_model/{name_or_id}", params={"name_or_id": name}))

    @classmethod
    def exists(cls, name_or_id: str) -> bool:
        """
        Check if a regression model exists in the OrcaCloud

        Params:
            name_or_id: Name or id of the regression model

        Returns:
            `True` if the regression model exists, `False` otherwise
        """
        try:
            cls.open(name_or_id)
            return True
        except LookupError:
            return False

    @classmethod
    def all(cls) -> list[RegressionModel]:
        """
        Get a list of handles to all regression models in the OrcaCloud

        Returns:
            List of handles to all regression models in the OrcaCloud
        """
        return [cls(metadata) for metadata in orca_api.GET("/regression_model")]

    @classmethod
    def drop(cls, name_or_id: str, if_not_exists: DropMode = "error"):
        """
        Delete a regression model from the OrcaCloud

        Warning:
            This will delete the model and all associated data, including predictions, evaluations, and feedback.

        Params:
            name_or_id: Name or id of the regression model
            if_not_exists: What to do if the regression model does not exist, defaults to `"error"`.
                Other option is `"ignore"` to do nothing if the regression model does not exist.

        Raises:
            LookupError: If the regression model does not exist and if_not_exists is `"error"`
        """
        try:
            orca_api.DELETE("/regression_model/{name_or_id}", params={"name_or_id": name_or_id})
            logging.info(f"Deleted model {name_or_id}")
        except LookupError:
            if if_not_exists == "error":
                raise

    def refresh(self):
        """Refresh the model data from the OrcaCloud"""
        self.__dict__.update(self.open(self.name).__dict__)

    def set(self, *, description: str | None = UNSET, locked: bool = UNSET) -> None:
        """
        Update editable attributes of the model.

        Note:
            If a field is not provided, it will default to [UNSET][orca_sdk.UNSET] and not be updated.

        Params:
            description: Value to set for the description
            locked: Value to set for the locked status

        Examples:
            Update the description:
            >>> model.set(description="New description")

            Remove description:
            >>> model.set(description=None)

            Lock the model:
            >>> model.set(locked=True)
        """
        update: PredictiveModelUpdate = {}
        if description is not UNSET:
            update["description"] = description
        if locked is not UNSET:
            update["locked"] = locked
        orca_api.PATCH("/regression_model/{name_or_id}", params={"name_or_id": self.id}, json=update)
        self.refresh()

    def lock(self) -> None:
        """Lock the model to prevent accidental deletion"""
        self.set(locked=True)

    def unlock(self) -> None:
        """Unlock the model to allow deletion"""
        self.set(locked=False)

    @overload
    def predict(
        self,
        value: str,
        expected_scores: float | None = None,
        tags: set[str] | None = None,
        save_telemetry: TelemetryMode = "on",
        prompt: str | None = None,
        use_lookup_cache: bool = True,
    ) -> RegressionPrediction: ...

    @overload
    def predict(
        self,
        value: list[str],
        expected_scores: list[float] | None = None,
        tags: set[str] | None = None,
        save_telemetry: TelemetryMode = "on",
        prompt: str | None = None,
        use_lookup_cache: bool = True,
    ) -> list[RegressionPrediction]: ...

    # TODO: add filter support
    def predict(
        self,
        value: str | list[str],
        expected_scores: float | list[float] | None = None,
        tags: set[str] | None = None,
        save_telemetry: TelemetryMode = "on",
        prompt: str | None = None,
        use_lookup_cache: bool = True,
    ) -> RegressionPrediction | list[RegressionPrediction]:
        """
        Make predictions using the regression model.

        Params:
            value: Input text(s) to predict scores for
            expected_scores: Expected score(s) for telemetry tracking
            tags: Tags to associate with the prediction(s)
            save_telemetry: Whether to save telemetry for the prediction(s), defaults to `True`,
                which will save telemetry asynchronously unless the `ORCA_SAVE_TELEMETRY_SYNCHRONOUSLY`
                environment variable is set to `"1"`. You can also pass `"sync"` or `"async"` to
                explicitly set the save mode.
            prompt: Optional prompt for instruction-tuned embedding models

        Returns:
            Single RegressionPrediction or list of RegressionPrediction objects

        Raises:
            ValueError: If expected_scores length doesn't match value length for batch predictions
        """
        telemetry_on, telemetry_sync = _get_telemetry_config(save_telemetry)
        response = orca_api.POST(
            "/gpu/regression_model/{name_or_id}/prediction",
            params={"name_or_id": self.id},
            json={
                "input_values": value if isinstance(value, list) else [value],
                "memoryset_override_name_or_id": self._memoryset_override_id,
                "expected_scores": (
                    expected_scores
                    if isinstance(expected_scores, list)
                    else [expected_scores] if expected_scores is not None else None
                ),
                "tags": list(tags or set()),
                "save_telemetry": telemetry_on,
                "save_telemetry_synchronously": telemetry_sync,
                "prompt": prompt,
                "use_lookup_cache": use_lookup_cache,
            },
        )

        if telemetry_on and any(p["prediction_id"] is None for p in response):
            raise RuntimeError("Failed to save prediction to database.")

        predictions = [
            RegressionPrediction(
                prediction_id=prediction["prediction_id"],
                label=None,
                label_name=None,
                score=prediction["score"],
                confidence=prediction["confidence"],
                anomaly_score=prediction["anomaly_score"],
                memoryset=self.memoryset,
                model=self,
                logits=None,
                input_value=input_value,
            )
            for prediction, input_value in zip(response, value if isinstance(value, list) else [value])
        ]
        self._last_prediction_was_batch = isinstance(value, list)
        self._last_prediction = predictions[-1]
        return predictions if isinstance(value, list) else predictions[0]

    def predictions(
        self,
        limit: int = 100,
        offset: int = 0,
        tag: str | None = None,
        sort: list[tuple[Literal["anomaly_score", "confidence", "timestamp"], Literal["asc", "desc"]]] = [],
    ) -> list[RegressionPrediction]:
        """
        Get a list of predictions made by this model

        Params:
            limit: Optional maximum number of predictions to return
            offset: Optional offset of the first prediction to return
            tag: Optional tag to filter predictions by
            sort: Optional list of columns and directions to sort the predictions by.
                Predictions can be sorted by `created_at`, `confidence`, `anomaly_score`, or `score`.

        Returns:
            List of score predictions

        Examples:
            Get the last 3 predictions:
            >>> predictions = model.predictions(limit=3, sort=[("created_at", "desc")])
            [
                RegressionPrediction({score: 4.5, confidence: 0.95, anomaly_score: 0.1, input_value: 'Great service'}),
                RegressionPrediction({score: 2.0, confidence: 0.90, anomaly_score: 0.1, input_value: 'Poor experience'}),
                RegressionPrediction({score: 3.5, confidence: 0.85, anomaly_score: 0.1, input_value: 'Average'}),
            ]

            Get second most confident prediction:
            >>> predictions = model.predictions(sort=[("confidence", "desc")], offset=1, limit=1)
            [RegressionPrediction({score: 4.2, confidence: 0.90, anomaly_score: 0.1, input_value: 'Good service'})]
        """
        predictions = orca_api.POST(
            "/telemetry/prediction",
            json={
                "model_id": self.id,
                "limit": limit,
                "offset": offset,
                "sort": [list(sort_item) for sort_item in sort],
                "tag": tag,
            },
        )
        return [
            RegressionPrediction(
                prediction_id=prediction["prediction_id"],
                label=None,
                label_name=None,
                score=prediction["score"],
                confidence=prediction["confidence"],
                anomaly_score=prediction["anomaly_score"],
                memoryset=self.memoryset,
                model=self,
                telemetry=prediction,
                logits=None,
                input_value=None,
            )
            for prediction in predictions
            if "score" in prediction
        ]

    def _evaluate_datasource(
        self,
        datasource: Datasource,
        value_column: str,
        score_column: str,
        record_predictions: bool,
        tags: set[str] | None,
        background: bool = False,
    ) -> RegressionMetrics | Job[RegressionMetrics]:
        response = orca_api.POST(
            "/regression_model/{model_name_or_id}/evaluation",
            params={"model_name_or_id": self.id},
            json={
                "datasource_name_or_id": datasource.id,
                "datasource_score_column": score_column,
                "datasource_value_column": value_column,
                "memoryset_override_name_or_id": self._memoryset_override_id,
                "record_telemetry": record_predictions,
                "telemetry_tags": list(tags) if tags else None,
            },
        )

        def get_value():
            res = orca_api.GET(
                "/regression_model/{model_name_or_id}/evaluation/{task_id}",
                params={"model_name_or_id": self.id, "task_id": response["task_id"]},
            )
            assert res["result"] is not None
            return RegressionMetrics(**res["result"])

        job = Job(response["task_id"], get_value)
        return job if background else job.result()

    def _evaluate_dataset(
        self,
        dataset: Dataset,
        value_column: str,
        score_column: str,
        record_predictions: bool,
        tags: set[str],
        batch_size: int,
        prompt: str | None = None,
    ) -> RegressionMetrics:
        predictions = [
            prediction
            for i in range(0, len(dataset), batch_size)
            for prediction in self.predict(
                dataset[i : i + batch_size][value_column],
                expected_scores=dataset[i : i + batch_size][score_column],
                tags=tags,
                save_telemetry="sync" if record_predictions else "off",
                prompt=prompt,
            )
        ]

        return calculate_regression_metrics(
            expected_scores=dataset[score_column],
            predicted_scores=[p.score for p in predictions],
            anomaly_scores=[p.anomaly_score for p in predictions],
        )

    @overload
    def evaluate(
        self,
        data: Datasource | Dataset,
        *,
        value_column: str = "value",
        score_column: str = "score",
        record_predictions: bool = False,
        tags: set[str] = {"evaluation"},
        batch_size: int = 100,
        prompt: str | None = None,
        background: Literal[True],
    ) -> Job[RegressionMetrics]:
        pass

    @overload
    def evaluate(
        self,
        data: Datasource | Dataset,
        *,
        value_column: str = "value",
        score_column: str = "score",
        record_predictions: bool = False,
        tags: set[str] = {"evaluation"},
        batch_size: int = 100,
        prompt: str | None = None,
        background: Literal[False] = False,
    ) -> RegressionMetrics:
        pass

    def evaluate(
        self,
        data: Datasource | Dataset,
        *,
        value_column: str = "value",
        score_column: str = "score",
        record_predictions: bool = False,
        tags: set[str] = {"evaluation"},
        batch_size: int = 100,
        prompt: str | None = None,
        background: bool = False,
    ) -> RegressionMetrics | Job[RegressionMetrics]:
        """
        Evaluate the regression model on a given dataset or datasource

        Params:
            data: Dataset or Datasource to evaluate the model on
            value_column: Name of the column that contains the input values to the model
            score_column: Name of the column containing the expected scores
            record_predictions: Whether to record [`RegressionPrediction`][orca_sdk.telemetry.RegressionPrediction]s for analysis
            tags: Optional tags to add to the recorded [`RegressionPrediction`][orca_sdk.telemetry.RegressionPrediction]s
            batch_size: Batch size for processing Dataset inputs (only used when input is a Dataset)
            prompt: Optional prompt for instruction-tuned embedding models
            background: Whether to run the operation in the background and return a job handle

        Returns:
            RegressionMetrics containing metrics including MAE, MSE, RMSE, R2, and anomaly score statistics

        Examples:
            >>> model.evaluate(datasource, value_column="text", score_column="rating")
            RegressionMetrics({
                mae: 0.2500,
                rmse: 0.3536,
                r2: 0.8500,
                anomaly_score: 0.3500 ± 0.0500,
            })

            >>> # Using with an instruction-tuned embedding model
            >>> model.evaluate(dataset,prompt="Represent this review for rating prediction:")
            RegressionMetrics({
            mae: 0.2000,
            rmse: 0.3000,
            r2: 0.9000,
            anomaly_score: 0.3000 ± 0.0400})
        """
        if isinstance(data, Datasource):
            return self._evaluate_datasource(
                datasource=data,
                value_column=value_column,
                score_column=score_column,
                record_predictions=record_predictions,
                tags=tags,
                background=background,
            )
        elif isinstance(data, Dataset):
            return self._evaluate_dataset(
                dataset=data,
                value_column=value_column,
                score_column=score_column,
                record_predictions=record_predictions,
                tags=tags,
                batch_size=batch_size,
                prompt=prompt,
            )
        else:
            raise ValueError(f"Invalid data type: {type(data)}")

    @contextmanager
    def use_memoryset(self, memoryset_override: ScoredMemoryset) -> Generator[None, None, None]:
        """
        Temporarily override the memoryset used by the model for predictions

        Params:
            memoryset_override: Memoryset to override the default memoryset with

        Examples:
            >>> with model.use_memoryset(ScoredMemoryset.open("my_other_memoryset")):
            ...     predictions = model.predict("Rate your experience")
        """
        self._memoryset_override_id = memoryset_override.id
        yield
        self._memoryset_override_id = None

    @overload
    def record_feedback(self, feedback: dict[str, Any]) -> None:
        pass

    @overload
    def record_feedback(self, feedback: Iterable[dict[str, Any]]) -> None:
        pass

    def record_feedback(self, feedback: Iterable[dict[str, Any]] | dict[str, Any]):
        """
        Record feedback for a list of predictions.

        We support recording feedback in several categories for each prediction. A
        [`FeedbackCategory`][orca_sdk.telemetry.FeedbackCategory] is created automatically,
        the first time feedback with a new name is recorded. Categories are global across models.
        The value type of the category is inferred from the first recorded value. Subsequent
        feedback for the same category must be of the same type.

        Params:
            feedback: Feedback to record, this should be dictionaries with the following keys:

                - `category`: Name of the category under which to record the feedback.
                - `value`: Feedback value to record, should be `True` for positive feedback and
                    `False` for negative feedback or a [`float`][float] between `-1.0` and `+1.0`
                    where negative values indicate negative feedback and positive values indicate
                    positive feedback.
                - `comment`: Optional comment to record with the feedback.

        Examples:
            Record whether predictions were accurate:
            >>> model.record_feedback({
            ...     "prediction": p.prediction_id,
            ...     "category": "accurate",
            ...     "value": abs(p.score - p.expected_score) < 0.5,
            ... } for p in predictions)

            Record star rating as normalized continuous score between `-1.0` and `+1.0`:
            >>> model.record_feedback({
            ...     "prediction": "123e4567-e89b-12d3-a456-426614174000",
            ...     "category": "rating",
            ...     "value": -0.5,
            ...     "comment": "2 stars"
            ... })

        Raises:
            ValueError: If the value does not match previous value types for the category, or is a
                [`float`][float] that is not between `-1.0` and `+1.0`.
        """
        orca_api.PUT(
            "/telemetry/prediction/feedback",
            json=[
                _parse_feedback(f) for f in (cast(list[dict], [feedback]) if isinstance(feedback, dict) else feedback)
            ],
        )
